##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def remapminmax(x, xmax, xmin):
    import numpy as np
    try:
        ndata, ninputs = np.shape(x)
        y = np.ones([ndata, ninputs])
    except Exception:
        ndata = np.size(x)
        ninputs = 1
        y = np.ones([ndata])
    if ninputs > 1:
        for j in range(ninputs):
            y[:, j] = (xmax[j] - xmin[j]) * (x[:, j] + 1) / (2.0) + xmin[j]
    else:
        y = (xmax - xmin) * (x + 1) / (2.0) + xmin
    return y



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Ideal gas + Ideal/Non-ideal liquid property package.

VLE calucations assuming an ideal gas for the gas phase. For the liquid phase,
options include ideal liquid or non-ideal liquid using an activity
coefficient model; options include Non Random Two Liquid Model (NRTL) or the
Wilson model to compute the activity coefficient. This property package
supports the following combinations for gas-liquid mixtures:
1. Ideal (vapor) - Ideal (liquid)
2. Ideal (vapor) - NRTL (liquid)
3. Ideal (vapor) - Wilson (liquid)

This property package currently supports the flow_mol, temperature, pressure
and mole_frac_comp as state variables (mole basis). Support for other
combinations will be available in the future.

Please note that the parameters required to compute the activity coefficient
for the component needs to be provided by the user in the parameter block or
can be estimated by the user if VLE data is available. Please see the
documentation for more details.

SI units.

References:

1. "The properties of gases and liquids by Robert C. Reid"
2. "Perry's Chemical Engineers Handbook by Robert H. Perry".
3. H. Renon and J.M. Prausnitz, "Local compositions in thermodynamic excess
   functions for liquid mixtures.", AIChE Journal Vol. 14, No.1, 1968.
"""

# Import Pyomo libraries
from pyomo.environ import Constraint, log, NonNegativeReals, value, Var, exp,\
    Set, Expression, Param, sqrt
from pyomo.opt import SolverFactory, TerminationCondition
from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            solve_indexed_blocks)
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

# Some more inforation about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.2"


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("ActivityCoeffParameterBlock")
class ActivityCoeffParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    Contains parameters and indexing sets associated with properties for
    BTX system.
    """
    # Config block for the _IdealStateBlock
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare("activity_coeff_model", ConfigValue(
        default="Ideal",
        domain=In(["Ideal", "NRTL", "Wilson"]),
        description="Flag indicating the activity coefficient model",
        doc="""Flag indicating the activity coefficient model to be used
for the non-ideal liquid, and thus corresponding constraints  should be
included,
**default** - Ideal liquid.
**Valid values:** {
**"NRTL"** - Non Random Two Liquid Model,
**"Wilson"** - Wilson Liquid Model,}"""))

    CONFIG.declare("state_vars", ConfigValue(
        default="FTPz",
        domain=In(["FTPz", "FcTP"]),
        description="Flag indicating the choice for state variables",
        doc="""Flag indicating the choice for state variables to be used
    for the state block, and thus corresponding constraints  should be
    included,
    **default** - FTPz
    **Valid values:** {
    **"FTPx"** - Total flow, Temperature, Pressure and Mole fraction,
    **"FcTP"** - Component flow, Temperature and Pressure}"""))

    CONFIG.declare("valid_phase", ConfigValue(
        default=("Vap", "Liq"),
        domain=In(["Liq", "Vap", ("Vap", "Liq"), ("Liq", "Vap")]),
        description="Flag indicating the valid phase",
        doc="""Flag indicating the valid phase for a given set of
conditions, and thus corresponding constraints  should be included,
**default** - ("Vap", "Liq").
**Valid values:** {
**"Liq"** - Liquid only,
**"Vap"** - Vapor only,
**("Vap", "Liq")** - Vapor-liquid equilibrium,
**("Liq", "Vap")** - Vapor-liquid equilibrium,}"""))

    def build(self):
        """Callable method for Block construction."""
        super(ActivityCoeffParameterData, self).build()

        self.state_block_class = ActivityCoeffStateBlock

        # List of valid phases in property package
        if self.config.valid_phase == ("Liq", "Vap") or \
                self.config.valid_phase == ("Vap", "Liq"):
            self.phase_list = Set(initialize=["Liq", "Vap"],
                                  ordered=True)
        elif self.config.valid_phase == "Liq":
            self.phase_list = Set(initialize=["Liq"])
        else:
            self.phase_list = Set(initialize=["Vap"])

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {"flow_mol": {"method": None, "units": "mol/s"},
             "mole_frac_comp": {"method": None, "units": "no unit"},
             "temperature": {"method": None, "units": "K"},
             "pressure": {"method": None, "units": "Pa"},
             "flow_mol_phase": {"method": None, "units": "mol/s"},
             "density_mol": {"method": "_density_mol",
                             "units": "mol/m^3"},
             "pressure_sat": {"method": "_pressure_sat", "units": "Pa"},
             "mole_frac_phase_comp": {"method": "_mole_frac_phase",
                                      "units": "no unit"},
             "energy_internal_mol_phase_comp": {
                 "method": "_energy_internal_mol_phase_comp",
                 "units": "J/mol"},
             "energy_internal_mol_phase": {
                 "method": "_energy_internal_mol_phase",
                 "units": "J/mol"},
             "enth_mol_phase_comp": {"method": "_enth_mol_phase_comp",
                                     "units": "J/mol"},
             "enth_mol_phase": {"method": "_enth_mol_phase",
                                "units": "J/mol"},
             "entr_mol_phase_comp": {"method": "_entr_mol_phase_comp",
                                     "units": "J/mol"},
             "entr_mol_phase": {"method": "_entr_mol_phase",
                                "units": "J/mol"},
             "gibbs_mol_phase_comp": {"method": "_gibbs_mol_phase_comp",
                                      "units": "J/mol"},
             "temperature_bubble": {"method": "_temperature_bubble",
                                    "units": "K"},
             "temperature_dew": {"method": "_temperature_dew",
                                 "units": "K"},
             "pressure_bubble": {"method": "_pressure_bubble",
                                 "units": "Pa"},
             "pressure_dew": {"method": "_pressure_dew",
                              "units": "Pa"},
             "fug_vap": {"method": "_fug_vap", "units": "Pa"},
             "fug_liq": {"method": "_fug_liq", "units": "Pa"},
             'ds_form': {'method': '_ds_form', 'units': 'J/mol.K'}})

        obj.add_default_units({"time": "s",
                               "length": "m",
                               "mass": "g",
                               "amount": "mol",
                               "temperature": "K",
                               "energy": "J",
                               "holdup": "mol"})


class _ActivityCoeffStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, state_args={}, hold_state=False,
                   state_vars_fixed=False, outlvl=5,
                   solver="ipopt", optarg={"tol": 1e-8}):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.

                         If FTPz are chosen as state_vars, then keys for
                         the state_args dictionary are:
                         flow_mol, temperature, pressure and mole_frac_comp

                         If FcTP are chose as the state_vars, then keys for
                         the state_args dictionary are:
                         flow_mol_comp, temperature, pressure.

            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = "ipopt")
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Deactivate the constraints specific for outlet block i.e.
        # when defined state is False
        init_log = getInitLogger(blk.name, outlvl)
        for k in blk.keys():
            if (blk[k].config.defined_state is False) and \
                    (blk[k]._params.config.state_vars == "FTPz"):
                blk[k].eq_mol_frac_out.deactivate()

        # Fix state variables if not already fixed
        if state_vars_fixed is False:
            flags = fix_state_vars(blk, state_args)
        else:
            # Check when the state vars are fixed already result in dof 0
            for k in blk.keys():
                if degrees_of_freedom(blk[k]) != 0:
                    raise Exception("State vars fixed but degrees of freedom "
                                    "for state block is not zero during "
                                    "initialization.")

        if optarg is None:
            sopt = {"tol": 1e-8}
        else:
            sopt = optarg

        opt = SolverFactory("ipopt")
        opt.options = sopt

        # ---------------------------------------------------------------------
        # Initialization sequence: Deactivating certain constraints
        # for 1st solve
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                if c.local_name in ["eq_total",
                                    "eq_comp",
                                    "eq_mole_frac"
                                    "eq_sum_mol_frac",
                                    "eq_phase_equilibrium",
                                    "eq_enth_mol_phase",
                                    "eq_entr_mol_phase",
                                    "eq_Gij_coeff",
                                    "eq_A",
                                    "eq_B",
                                    "eq_activity_coeff"]:
                    c.deactivate()

        # First solve for the active constraints that remain (p_sat, T_bubble,
        # T_dew). Valid only for a 2 phase block. If single phase,
        # no constraints are active.
        # NOTE: "k" is the last value from the previous for loop
        # only for the purpose of having a valid index. The assumption
        # here is that for all values of "k", the attribute exists.
        if (blk[k].config.has_phase_equilibrium) or \
                (blk[k].config.parameters.config.valid_phase ==
                    ("Liq", "Vap")) or \
                (blk[k].config.parameters.config.valid_phase ==
                    ("Vap", "Liq")):
            results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
            init_log.log(4, "Initialization Step 1 {}.".format(condition(results)))
        else:
            init_log.info(4, "Initialization step 1 skipped")

        # Continue initialization sequence and activate select constraints
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                if c.local_name in ["eq_total",
                                    "eq_comp",
                                    "eq_sum_mol_frac",
                                    "eq_phase_equilibrium"]:
                    c.activate()
            if blk[k].config.parameters.config.activity_coeff_model \
                    != "Ideal":
                # assume ideal and solve
                blk[k].activity_coeff_comp.fix(1)

        # Second solve for the active constraints
        results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        init_log.log(4, "Initialization Step 2 {}.".format(condition(results)))

        # Activate activity coefficient specific constraints
        for k in blk.keys():
            if blk[k].config.parameters.config.activity_coeff_model \
                    != "Ideal":
                for c in blk[k].component_objects(Constraint):
                    if c.local_name in ["eq_Gij_coeff",
                                        "eq_A",
                                        "eq_B"]:
                        c.activate()

        results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        init_log.log(4, "Initialization Step 3 {}.".format(condition(results)))

        for k in blk.keys():
            if blk[k].config.parameters.config.activity_coeff_model \
                    != "Ideal":
                blk[k].eq_activity_coeff.activate()
                blk[k].activity_coeff_comp.unfix()

        results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        init_log.log(4, "Initialization Step 4 {}.".format(condition(results)))

        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                if c.local_name in ["eq_enth_mol_phase",
                                    "eq_entr_mol_phase"]:
                    c.activate()

        results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        init_log.log(4, "Initialization Step 5 {}.".format(condition(results)))

        for k in blk.keys():
            if (blk[k].config.defined_state is False) and \
                    (blk[k]._params.config.state_vars == "FTPz"):
                blk[k].eq_mol_frac_out.activate()

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

        init_log.log(5, "Initialization Complete: {}".format(condition(results)))


    def release_state(blk, flags, outlvl=6):
        """
        Method to relase state variables fixed during initialization.
        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        if flags is None:
            return

        # Unfix state variables
        revert_state_vars(blk, flags)

        if outlvl > 0:
            if outlvl > 0:
                _log.info("{} State Released.".format(blk.name))


@declare_process_block_class("ActivityCoeffStateBlock",
                             block_class=_ActivityCoeffStateBlock)
class ActivityCoeffStateBlockData(StateBlockData):
    """An example property package for ideal VLE."""

    def build(self):
        """Callable method for Block construction."""
        super(ActivityCoeffStateBlockData, self).build()

        # Check for valid phase indicator and consistent flags
        if self.config.has_phase_equilibrium and \
                self.config.parameters.config.valid_phase in ["Vap", "Liq"]:
            raise ConfigurationError("Inconsistent inputs. Valid phase"
                                     " flag not set to VL for the state"
                                     " block but has_phase_equilibrium"
                                     " is set to True.")
        self._make_state_vars()
        self._make_vars()
        if not self.config.has_phase_equilibrium and \
                self.config.parameters.config.valid_phase == "Liq":
            self._make_liq_phase_eq()

        if (self.config.has_phase_equilibrium) or \
                (self.config.parameters.config.valid_phase ==
                    ("Liq", "Vap")) or \
                (self.config.parameters.config.valid_phase ==
                    ("Vap", "Liq")):
            if self.config.parameters.config.activity_coeff_model == "NRTL":
                self._make_NRTL_eq()
            if self.config.parameters.config.activity_coeff_model == "Wilson":
                self._make_Wilson_eq()
            self._make_flash_eq()

        if not self.config.has_phase_equilibrium and \
                self.config.parameters.config.valid_phase == "Vap":
            self._make_vap_phase_eq()

    def _make_state_vars(self):
        """List the necessary state variable objects."""

        if self._params.config.state_vars == "FTPz":
            self.flow_mol = Var(initialize=1.0,
                                domain=NonNegativeReals,
                                doc="Total molar flowrate [mol/s]")
            self.mole_frac_comp = Var(
                    self._params.component_list,
                    bounds=(0, 1),
                    initialize=1 / len(self._params.component_list),
                    doc="Mixture mole fraction")
            self.pressure = Var(initialize=101325,
                                domain=NonNegativeReals,
                                doc="State pressure [Pa]")
            self.temperature = Var(initialize=298.15,
                                   domain=NonNegativeReals,
                                   doc="State temperature [K]")
        else:
            self.flow_mol_comp = Var(
                    self._params.component_list,
                    initialize=1 / len(self._params.component_list),
                    domain=NonNegativeReals,
                    doc="Component molar flowrate [mol/s]")
            self.pressure = Var(initialize=101325,
                                domain=NonNegativeReals,
                                doc="State pressure [Pa]")
            self.temperature = Var(initialize=298.15,
                                   domain=NonNegativeReals,
                                   doc="State temperature [K]")

    def _make_vars(self):

        if self._params.config.state_vars == "FTPz":
            self.flow_mol_phase = Var(self._params.phase_list,
                                      initialize=0.5)
        else:
            self.flow_mol_phase_comp = Var(self._params.phase_list,
                                           self._params.component_list,
                                           initialize=0.5)

            def rule_mix_mole_frac(self, i):
                return self.flow_mol_comp[i] / \
                    sum(self.flow_mol_comp[i]
                        for i in self._params.component_list)
            self.mole_frac_comp = Expression(self._params.component_list,
                                             rule=rule_mix_mole_frac)

        self.mole_frac_phase_comp = \
            Var(self._params.phase_list,
                self._params.component_list,
                initialize=1 / len(self._params.component_list),
                bounds=(0, 1))

    def _make_liq_phase_eq(self):

        if self._params.config.state_vars == "FTPz":
            def rule_total_mass_balance(self):
                return self.flow_mol_phase["Liq"] == self.flow_mol
            self.eq_total = Constraint(rule=rule_total_mass_balance)

            def rule_comp_mass_balance(self, i):
                return self.flow_mol * self.mole_frac_comp[i] == \
                    self.flow_mol_phase["Liq"] * \
                    self.mole_frac_phase_comp["Liq", i]
            self.eq_comp = Constraint(self._params.component_list,
                                      rule=rule_comp_mass_balance)

            if self.config.defined_state is False:
                # applied at outlet only
                self.eq_mol_frac_out = \
                    Constraint(expr=sum(self.mole_frac_comp[i]
                               for i in self._params.component_list) == 1)
        else:
            def rule_comp_mass_balance(self, i):
                return self.flow_mol_comp[i] == \
                    self.flow_mol_phase_comp["Liq", i]
            self.eq_comp = Constraint(self._params.component_list,
                                      rule=rule_comp_mass_balance)

            def rule_mole_frac(self, i):
                return self.mole_frac_phase_comp["Liq", i] * \
                    sum(self.flow_mol_phase_comp["Liq", i]
                        for i in self._params.component_list) == \
                    self.flow_mol_phase_comp["Liq", i]
            self.eq_mole_frac = Constraint(self._params.component_list,
                                           rule=rule_mole_frac)

    def _make_vap_phase_eq(self):

        if self._params.config.state_vars == "FTPz":
            def rule_total_mass_balance(self):
                return self.flow_mol_phase["Vap"] == self.flow_mol
            self.eq_total = Constraint(rule=rule_total_mass_balance)

            def rule_comp_mass_balance(self, i):
                return self.flow_mol * self.mole_frac_comp[i] == \
                    self.flow_mol_phase["Vap"] * \
                    self.mole_frac_phase_comp["Vap", i]
            self.eq_comp = Constraint(self._params.component_list,
                                      rule=rule_comp_mass_balance)

            if self.config.defined_state is False:
                # applied at outlet only
                self.eq_mol_frac_out = \
                    Constraint(expr=sum(self.mole_frac_comp[i]
                                        for i in self._params.component_list)
                               == 1)
        else:
            def rule_comp_mass_balance(self, i):
                return self.flow_mol_comp[i] == \
                    self.flow_mol_phase_comp["Vap", i]
            self.eq_comp = Constraint(self._params.component_list,
                                      rule=rule_comp_mass_balance)

            def rule_mole_frac(self, i):
                return self.mole_frac_phase_comp["Vap", i] * \
                    sum(self.flow_mol_phase_comp["Vap", i]
                        for i in self._params.component_list) == \
                    self.flow_mol_phase_comp["Vap", i]
            self.eq_mole_frac = Constraint(self._params.component_list,
                                           rule=rule_mole_frac)

    def _make_flash_eq(self):

        if self._params.config.state_vars == "FTPz":
            # Total mole balance
            def rule_total_mass_balance(self):
                return self.flow_mol_phase["Liq"] + \
                    self.flow_mol_phase["Vap"] == self.flow_mol
            self.eq_total = Constraint(rule=rule_total_mass_balance)

            # Component mole balance
            def rule_comp_mass_balance(self, i):
                return self.flow_mol * self.mole_frac_comp[i] == (
                    self.flow_mol_phase["Liq"] *
                    self.mole_frac_phase_comp["Liq", i] +
                    self.flow_mol_phase["Vap"] *
                    self.mole_frac_phase_comp["Vap", i])
            self.eq_comp = Constraint(self._params.component_list,
                                      rule=rule_comp_mass_balance)

            # sum of mole fractions constraint (sum(x_i)-sum(y_i)=0)
            def rule_mole_frac(self):
                return sum(self.mole_frac_phase_comp["Liq", i]
                           for i in self._params.component_list) -\
                    sum(self.mole_frac_phase_comp["Vap", i]
                        for i in self._params.component_list) == 0
            self.eq_sum_mol_frac = Constraint(rule=rule_mole_frac)

            if self.config.defined_state is False:
                # applied at outlet only as complete state information unknown
                self.eq_mol_frac_out = \
                    Constraint(expr=sum(self.mole_frac_comp[i]
                               for i in self._params.component_list) == 1)
        else:
            def rule_comp_mass_balance(self, i):
                return self.flow_mol_comp[i] == \
                    self.flow_mol_phase_comp["Liq", i] + \
                    self.flow_mol_phase_comp["Vap", i]
            self.eq_comp = Constraint(self._params.component_list,
                                      rule=rule_comp_mass_balance)

            def rule_mole_frac(self, p, i):
                return self.mole_frac_phase_comp[p, i] * \
                    sum(self.flow_mol_phase_comp[p, i]
                        for i in self._params.component_list) == \
                    self.flow_mol_phase_comp[p, i]
            self.eq_mole_frac = Constraint(self._params.phase_list,
                                           self._params.component_list,
                                           rule=rule_mole_frac)

        # Smooth Flash Formulation

        # Please refer to Burgard et al., "A Smooth, Square Flash
        # Formulation for Equation Oriented Flowsheet Optimization",
        # Computer Aided Chemical Engineering 44, 871-876, 2018.

        self._temperature_equilibrium = \
            Var(initialize=self.temperature.value,
                doc="Temperature for calculating "
                    "phase equilibrium")

        self._t1 = Var(initialize=self.temperature.value,
                       doc="Intermediate temperature for calculating "
                           "the equilibrium temperature")

        self.eps_1 = Param(default=0.01,
                           mutable=True,
                           doc="Smoothing parameter for equilibrium "
                               "temperature")
        self.eps_2 = Param(default=0.0005,
                           mutable=True,
                           doc="Smoothing parameter for equilibrium "
                               "temperature")

        # Equation #13 in reference cited above
        # Approximation for max(temperature, temperature_bubble)
        def rule_t1(b):
            return b._t1 == 0.5 * \
                (b.temperature + b.temperature_bubble +
                 sqrt((b.temperature - b.temperature_bubble)**2 +
                      b.eps_1**2))
        self._t1_constraint = Constraint(rule=rule_t1)

        # Equation #14 in reference cited above
        # Approximation for min(_t1, temperature_dew)
        # TODO : Add option for supercritical extension
        def rule_teq(b):
            return b._temperature_equilibrium == 0.5 * \
                (b._t1 + b.temperature_dew -
                 sqrt((b._t1 - b.temperature_dew)**2 +
                      b.eps_2**2))
        self._teq_constraint = Constraint(rule=rule_teq)

        def rule_phase_eq(self, i):
            return self.fug_vap[i] == self.fug_liq[i]
        self.eq_phase_equilibrium = Constraint(self._params.component_list,
                                               rule=rule_phase_eq)

    def _make_NRTL_eq(self):

        # NRTL Model specific variables (values to be fixed by user or need to
        # be estimated based on VLE data)
        # See documentation for suggested or typical values.
        self.alpha = Var(self._params.component_list,
                         self._params.component_list,
                         initialize=0.3,
                         doc="Non-randomness parameter for NRTL model")

        self.tau = Var(self._params.component_list,
                       self._params.component_list,
                       initialize=1.0,
                       doc="Binary interaction parameter for NRTL model")

        # NRTL model variables
        self.Gij_coeff = Var(self._params.component_list,
                             self._params.component_list,
                             initialize=1.0,
                             doc="Gij coefficient for use in NRTL model ")

        self.activity_coeff_comp = Var(self._params.component_list,
                                       initialize=1.0,
                                       doc="Activity coefficient of component")

        self.A = Var(self._params.component_list,
                     initialize=1.0,
                     doc="Intermediate variable to compute activity"
                     " coefficient")

        self.B = Var(self._params.component_list,
                     initialize=1.0,
                     doc="Intermediate variable to compute activity"
                     " coefficient")

        def rule_Gij_coeff(self, i, j):
            # i,j component
            if i != j:
                return self.Gij_coeff[i, j] == exp(-self.alpha[i, j] *
                                                   self.tau[i, j])
            else:
                self.Gij_coeff[i, j].fix(1)
                return Constraint.Skip

        self.eq_Gij_coeff = Constraint(self._params.component_list,
                                       self._params.component_list,
                                       rule=rule_Gij_coeff)

        # First sum part in the NRTL equation
        def rule_A(self, i):
            value_1 = sum(self.mole_frac_phase_comp["Liq", j] *
                          self.tau[j, i] * self.Gij_coeff[j, i]
                          for j in self._params.component_list)
            value_2 = sum(self.mole_frac_phase_comp["Liq", k] *
                          self.Gij_coeff[k, i]
                          for k in self._params.component_list)
            return self.A[i] == value_1 / value_2
        self.eq_A = Constraint(self._params.component_list, rule=rule_A)

        # Second sum part in the NRTL equation
        def rule_B(self, i):
            value = sum((self.mole_frac_phase_comp["Liq", j] *
                         self.Gij_coeff[i, j] /
                         sum(self.mole_frac_phase_comp["Liq", k] *
                             self.Gij_coeff[k, j]
                             for k in self._params.component_list)) *
                        (self.tau[i, j] - sum(
                                self.mole_frac_phase_comp["Liq", m] *
                                self.tau[m, j] *
                                self.Gij_coeff[m, j]
                                for m in self._params.component_list) /
                         sum(self.mole_frac_phase_comp["Liq", k] *
                         self.Gij_coeff[k, j]
                         for k in self._params.component_list))
                        for j in self._params.component_list)
            return self.B[i] == value
        self.eq_B = Constraint(self._params.component_list, rule=rule_B)

        # Activity coefficient using NRTL
        def rule_activity_coeff(self, i):
            return log(self.activity_coeff_comp[i]) == self.A[i] + self.B[i]
        self.eq_activity_coeff = Constraint(self._params.component_list,
                                            rule=rule_activity_coeff)

    def _make_Wilson_eq(self):

        # Wilson Model specific variables (values to be fixed by user or need
        # to be estimated based on VLE data)
        self.vol_mol_comp = Var(self._params.component_list,
                                initialize=1.0,
                                doc="Molar volume of component")

        self.tau = Var(self._params.component_list,
                       self._params.component_list,
                       initialize=1.0,
                       doc="Binary interaction parameter for Wilson model")

        # Wilson model variables
        self.Gij_coeff = Var(self._params.component_list,
                             self._params.component_list,
                             initialize=1.0,
                             doc="Gij coefficient for use in Wilson model ")

        self.activity_coeff_comp = Var(self._params.component_list,
                                       initialize=1.0,
                                       doc="Activity coefficient of component")

        self.A = Var(self._params.component_list,
                     initialize=1.0,
                     doc="Intermediate variable to compute activity"
                     " coefficient")

        self.B = Var(self._params.component_list,
                     initialize=1.0,
                     doc="Intermediate variable to compute activity"
                     " coefficient")

        def rule_Gij_coeff(self, i, j):
            # component i,j
            if i != j:
                return self.Gij_coeff[i, j] == \
                    (self.vol_mol_comp[i] /
                     self.vol_mol_comp[j]) * exp(-self.tau[i, j])
            else:
                self.Gij_coeff[i, j].fix(1)
                return Constraint.Skip

        self.eq_Gij_coeff = Constraint(self._params.component_list,
                                       self._params.component_list,
                                       rule=rule_Gij_coeff)

        # First sum part in Wilson equation
        def rule_A(self, i):
            value_1 = log(sum(self.mole_frac_phase_comp["Liq", j] *
                              self.Gij_coeff[j, i]
                              for j in self._params.component_list))
            return self.A[i] == value_1
        self.eq_A = Constraint(self._params.component_list, rule=rule_A)

        # Second sum part in Wilson equation
        def rule_B(self, i):
            value = sum((self.mole_frac_phase_comp["Liq", j] *
                         self.Gij_coeff[i, j] /
                         sum(self.mole_frac_phase_comp["Liq", k] *
                             self.Gij_coeff[k, j]
                             for k in self._params.component_list))
                        for j in self._params.component_list)
            return self.B[i] == value
        self.eq_B = Constraint(self._params.component_list, rule=rule_B)

        # Activity coefficient using Wilson equation
        def rule_activity_coeff(self, i):
            return log(self.activity_coeff_comp[i]) == \
                1 - self.A[i] - self.B[i]
        self.eq_activity_coeff = Constraint(self._params.component_list,
                                            rule=rule_activity_coeff)

    def _pressure_sat(self):
        self.pressure_sat = Var(self._params.component_list,
                                initialize=101325,
                                doc="vapor pressure ")

        def rule_reduced_temp(self, i):
            # reduced temperature is variable "x" in the documentation
            return (self._params.temperature_critical[i] -
                    self._temperature_equilibrium) / \
                self._params.temperature_critical[i]
        self._reduced_temp = Expression(self._params.component_list,
                                        rule=rule_reduced_temp)

        def rule_P_vap(self, j):
            return (1 - self._reduced_temp[j]) * \
                log(self.pressure_sat[j] /
                    self._params.pressure_critical[j]) == \
                (self._params.pressure_sat_coeff[j, "A"] *
                 self._reduced_temp[j] +
                 self._params.pressure_sat_coeff[j, "B"] *
                 self._reduced_temp[j]**1.5 +
                 self._params.pressure_sat_coeff[j, "C"] *
                 self._reduced_temp[j]**3 +
                 self._params.pressure_sat_coeff[j, "D"] *
                 self._reduced_temp[j]**6)
        self.eq_P_vap = Constraint(self._params.component_list,
                                   rule=rule_P_vap)

    def _fug_vap(self):
        def rule_fug_vap(self, i):
            return self.mole_frac_phase_comp["Vap", i] * self.pressure
        self.fug_vap = Expression(self._params.component_list,
                                  rule=rule_fug_vap)

    def _fug_liq(self):
        def rule_fug_liq(self, i):
            if self.config.parameters.config.\
                    activity_coeff_model == "Ideal":
                return self.mole_frac_phase_comp["Liq", i] * \
                    self.pressure_sat[i]
            else:
                return self.mole_frac_phase_comp["Liq", i] * \
                    self.activity_coeff_comp[i] * self.pressure_sat[i]
        self.fug_liq = Expression(self._params.component_list,
                                  rule=rule_fug_liq)

    def _density_mol(self):
        self.density_mol = Var(self._params.phase_list, doc="Molar density")

        def density_mol_calculation(self, p):
            if p == "Vap":
                return self.pressure == (self.density_mol[p] *
                                         self._params.gas_const *
                                         self.temperature)
            elif p == "Liq":  # TODO: Add a correlation to compute liq density
                _log.warning("Using a place holder for liquid density "
                             "{}. Please provide value or expression to "
                             "compute the liquid density".format(self.name))
                return self.density_mol[p] == 11.1E3  # mol/m3

        try:
            # Try to build constraint
            self.density_mol_calculation = Constraint(
                self._params.phase_list, rule=density_mol_calculation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.density_mol)
            self.del_component(self.density_mol_calculation)
            raise

    def _energy_internal_mol_phase(self):
        self.energy_internal_mol_phase = Var(
            self._params.phase_list,
            doc="Phase molar specific internal energy [J/mol]")

        def rule_energy_internal_mol_phase(b, p):
            return b.energy_internal_mol_phase[p] == sum(
                b.energy_internal_mol_phase_comp[p, i] *
                b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list)
        self.eq_energy_internal_mol_phase = Constraint(
            self._params.phase_list,
            rule=rule_energy_internal_mol_phase)

    def _energy_internal_mol_phase_comp(self):
        self.energy_internal_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            doc="Phase-component molar specific internal energy [J/mol]")

        def rule_energy_internal_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b.energy_internal_mol_phase_comp[p, j] == \
                    b.enth_mol_phase_comp[p, j] - \
                    b._params.gas_const * (b.temperature -
                                           b._params.temperature_ref)
            else:
                return b.energy_internal_mol_phase_comp[p, j] == \
                    b.enth_mol_phase_comp[p, j]
        self.eq_energy_internal_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_energy_internal_mol_phase_comp)

    def _enth_mol_phase(self):
        self.enth_mol_phase = Var(
            self._params.phase_list,
            doc="Phase molar specific enthalpies [J/mol]")

        def rule_enth_mol_phase(b, p):
            return b.enth_mol_phase[p] == sum(
                b.enth_mol_phase_comp[p, i] *
                b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list)
        self.eq_enth_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_enth_mol_phase)

    def _enth_mol_phase_comp(self):
        self.enth_mol_phase_comp = Var(self._params.phase_list,
                                       self._params.component_list,
                                       doc="Phase-component molar specific "
                                           "enthalpies [J/mol]")

        def rule_enth_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b._enth_mol_comp_vap(j)
            else:
                return b._enth_mol_comp_liq(j)
        self.eq_enth_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_enth_mol_phase_comp)

    def _enth_mol_comp_liq(self, j):
        # Liquid phase comp enthalpy (J/mol)
        # 1E3 conversion factor to convert from J/kmol to J/mol
        return self.enth_mol_phase_comp["Liq", j] * 1E3 == \
            1e3*self._params.dh_form["Liq", j] + \
            ((self._params.CpIG["Liq", j, "E"] / 5) *
                (self.temperature**5 -
                 self._params.temperature_reference**5)
                + (self._params.CpIG["Liq", j, "D"] / 4) *
                  (self.temperature**4 -
                   self._params.temperature_reference**4)
                + (self._params.CpIG["Liq", j, "C"] / 3) *
                  (self.temperature**3 -
                   self._params.temperature_reference**3)
                + (self._params.CpIG["Liq", j, "B"] / 2) *
                  (self.temperature**2 -
                   self._params.temperature_reference**2)
                + self._params.CpIG["Liq", j, "A"] *
                  (self.temperature - self._params.temperature_reference))

    def _enth_mol_comp_vap(self, j):

        # Vapor phase component enthalpy (J/mol)
        return self.enth_mol_phase_comp["Vap", j] == \
            self._params.dh_form["Vap", j] + \
            ((self._params.CpIG["Vap", j, "E"] / 5) *
                (self.temperature**5 -
                 self._params.temperature_reference**5)
                + (self._params.CpIG["Vap", j, "D"] / 4) *
                  (self.temperature**4 -
                   self._params.temperature_reference**4)
                + (self._params.CpIG["Vap", j, "C"] / 3) *
                  (self.temperature**3 -
                   self._params.temperature_reference**3)
                + (self._params.CpIG["Vap", j, "B"] / 2) *
                  (self.temperature**2 -
                   self._params.temperature_reference**2)
                + self._params.CpIG["Vap", j, "A"] *
                  (self.temperature -
                   self._params.temperature_reference))

    def _entr_mol_phase(self):
        self.entr_mol_phase = Var(
            self._params.phase_list,
            doc="Phase molar specific enthropies [J/mol.K]")

        def rule_entr_mol_phase(self, p):
            return self.entr_mol_phase[p] == sum(
                self.entr_mol_phase_comp[p, i] *
                self.mole_frac_phase_comp[p, i]
                for i in self._params.component_list)
        self.eq_entr_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_entr_mol_phase)

    def _entr_mol_phase_comp(self):
        self.entr_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            doc="Phase-component molar specific entropies [J/mol.K]")

        def rule_entr_mol_phase_comp(self, p, j):
            if p == "Vap":
                return self._entr_mol_comp_vap(j)
            else:
                return self._entr_mol_comp_liq(j)
        self.eq_entr_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_entr_mol_phase_comp)

    def _entr_mol_comp_liq(self, j):
        # Liquid phase comp entropy (J/mol.K)
        # 1E3 conversion factor to convert from J/kmol.K to J/mol.K
        return self.entr_mol_phase_comp['Liq', j] * 1E3 == \
            1E3*self._params.ds_form["Liq", j] + (
            ((self._params.CpIG['Liq', j, 'E'] / 4) *
                (self.temperature**4 - self._params.temperature_reference**4)
                + (self._params.CpIG['Liq', j, 'D'] / 3) *
                  (self.temperature**3 - self._params.temperature_reference**3)
                + (self._params.CpIG['Liq', j, 'C'] / 2) *
                  (self.temperature**2 - self._params.temperature_reference**2)
                + self._params.CpIG['Liq', j, 'B'] *
                  (self.temperature - self._params.temperature_reference)
                + self._params.CpIG['Liq', j, 'A'] *
             log(self.temperature / self._params.temperature_reference)))

    def _entr_mol_comp_vap(self, j):
        # component molar entropy of vapor phase
        return self.entr_mol_phase_comp["Vap", j] == (
            self._params.ds_form["Vap", j] +
            ((self._params.CpIG['Vap', j, 'E'] / 4) *
             (self.temperature**4 - self._params.temperature_reference**4)
             + (self._params.CpIG['Vap', j, 'D'] / 3) *
             (self.temperature**3 - self._params.temperature_reference**3)
             + (self._params.CpIG['Vap', j, 'C'] / 2) *
               (self.temperature**2 - self._params.temperature_reference**2)
                + self._params.CpIG['Vap', j, 'B'] *
               (self.temperature - self._params.temperature_reference)
                + self._params.CpIG['Vap', j, 'A'] *
                log(self.temperature / self._params.temperature_reference)) -
            self._params.gas_const * log(self.mole_frac_phase_comp['Vap', j] *
                                         self.pressure /
                                         self._params.pressure_reference))

    def _gibbs_mol_phase_comp(self):
        self.gibbs_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            doc="Phase-component molar specific Gibbs energies [J/mol]")

        def rule_gibbs_mol_phase_comp(self, p, j):
            return self.gibbs_mol_phase_comp[p, j] == \
                    self.enth_mol_phase_comp[p, j] - \
                    self.temperature*self.entr_mol_phase_comp[p, j]
        self.eq_gibbs_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_gibbs_mol_phase_comp)

    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        if (p == "Vap") and (j in self._params.component_list):
            if self._params.config.state_vars == "FTPz":
                return self.flow_mol_phase["Vap"] * \
                    self.mole_frac_phase_comp["Vap", j]
            else:
                return self.flow_mol_phase_comp["Vap", j]
        elif (p == "Liq") and (j in self._params.component_list):
            if self._params.config.state_vars == "FTPz":
                return self.flow_mol_phase["Liq"] * \
                    self.mole_frac_phase_comp["Liq", j]
            else:
                return self.flow_mol_phase_comp["Liq", j]
        else:
            return 0

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        if p == "Vap":
            if self._params.config.state_vars == "FTPz":
                return self.flow_mol_phase["Vap"] * self.enth_mol_phase["Vap"]
            else:
                return sum(self.flow_mol_phase_comp["Vap", j]
                           for j in self._params.component_list) * \
                    self.enth_mol_phase["Vap"]
        elif p == "Liq":
            if self._params.config.state_vars == "FTPz":
                return self.flow_mol_phase["Liq"] * self.enth_mol_phase["Liq"]
            else:
                return sum(self.flow_mol_phase_comp["Liq", j]
                           for j in self._params.component_list) * \
                    self.enth_mol_phase["Liq"]

    def get_material_density_terms(self, p, j):
        """Create material density terms."""
        if p == "Liq":
            if j in self._params.component_list:
                return self.density_mol[p]*self.mole_frac_phase_comp["Liq", j]
            else:
                return 0
        elif p == "Vap":
            if j in self._params.component_list:
                return self.density_mol[p]*self.mole_frac_phase_comp["Vap", j]
            else:
                return 0

    def get_energy_density_terms(self, p):
        """Create enthalpy density terms."""
        if p == "Liq":
            return self.density_mol[p] * self.energy_internal_mol_phase["Liq"]
        elif p == "Vap":
            return self.density_mol[p] * self.energy_internal_mol_phase["Vap"]

    def get_material_flow_basis(self):
        """Declare material flow basis."""
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        """Define state vars."""

        if self._params.config.state_vars == "FTPz":
            return {"flow_mol": self.flow_mol,
                    "mole_frac_comp": self.mole_frac_comp,
                    "temperature": self.temperature,
                    "pressure": self.pressure}
        else:
            return {"flow_mol_comp": self.flow_mol_comp,
                    "temperature": self.temperature,
                    "pressure": self.pressure}

    def model_check(blk):
        """Model checks for property block."""
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error("{} Temperature set below lower bound."
                       .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error("{} Temperature set above upper bound."
                       .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error("{} Pressure set below lower bound.".format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error("{} Pressure set above upper bound.".format(blk.name))

# -----------------------------------------------------------------------------
# Bubble and Dew Points
    def _temperature_bubble(self):
        self.temperature_bubble = Var(initialize=298.15,
                                      doc="Bubble point temperature (K)")

        def rule_psat_bubble(m, j):
            return self._params.pressure_critical[j] * \
                exp((self._params.pressure_sat_coeff[j, "A"] *
                    (1 - self.temperature_bubble /
                    self._params.temperature_critical[j]) +
                    self._params.pressure_sat_coeff[j, "B"] *
                    (1 - self.temperature_bubble /
                    self._params.temperature_critical[j])**1.5 +
                    self._params.pressure_sat_coeff[j, "C"] *
                    (1 - self.temperature_bubble /
                    self._params.temperature_critical[j])**3 +
                    self._params.pressure_sat_coeff[j, "D"] *
                    (1 - self.temperature_bubble /
                    self._params.temperature_critical[j])**6) /
                    (1 - (1 - self.temperature_bubble /
                          self._params.temperature_critical[j])))

        try:
            # Try to build expression
            self._p_sat_bubbleT = Expression(self._params.component_list,
                                             rule=rule_psat_bubble)

            def rule_temp_bubble(self):
                if self.config.parameters.config.activity_coeff_model == \
                        "Ideal":

                    return sum(self.mole_frac_comp[i] *
                               self._p_sat_bubbleT[i]
                               for i in self._params.component_list) - \
                        self.pressure == 0
                elif self.config.parameters.config.\
                        activity_coeff_model == "NRTL":
                    # NRTL model variables
                    def rule_Gij_coeff_bubble(self, i, j):
                        if i != j:
                            return exp(-self.alpha[i, j] * self.tau[i, j])
                        else:
                            return 1

                    self.Gij_coeff_bubble = Expression(
                        self._params.component_list,
                        self._params.component_list,
                        rule=rule_Gij_coeff_bubble)

                    def rule_A_bubble(self, i):
                        value_1 = sum(self.mole_frac_comp[j] *
                                      self.tau[j, i] *
                                      self.Gij_coeff_bubble[j, i]
                                      for j in self._params.component_list)
                        value_2 = sum(self.mole_frac_comp[k] *
                                      self.Gij_coeff_bubble[k, i]
                                      for k in self._params.component_list)
                        return value_1 / value_2
                    self.A_bubble = Expression(self._params.component_list,
                                               rule=rule_A_bubble)

                    def rule_B_bubble(self, i):
                        value = sum(
                                (self.mole_frac_comp[j] *
                                 self.Gij_coeff_bubble[i, j] /
                                 sum(self.mole_frac_comp[k] *
                                     self.Gij_coeff_bubble[k, j]
                                     for k in self._params.component_list)) *
                                (self.tau[i, j] - sum(
                                    self.mole_frac_comp[m] *
                                    self.tau[m, j] *
                                    self.Gij_coeff_bubble[m, j]
                                    for m in self._params.component_list) /
                                 sum(self.mole_frac_comp[k] *
                                     self.Gij_coeff_bubble[k, j]
                                     for k in self._params.component_list))
                                for j in self._params.component_list)
                        return value
                    self.B_bubble = Expression(self._params.component_list,
                                               rule=rule_B_bubble)

                    def rule_activity_coeff_bubble(self, i):
                        return exp(self.A_bubble[i] + self.B_bubble[i])
                    self.activity_coeff_comp_bubble = \
                        Expression(self._params.component_list,
                                   rule=rule_activity_coeff_bubble)

                    return sum(self.mole_frac_comp[i] *
                               self.activity_coeff_comp_bubble[i] *
                               self._p_sat_bubbleT[i]
                               for i in self._params.component_list) - \
                        self.pressure == 0
                else:
                    def rule_Gij_coeff_bubble(self, i, j):
                        if i != j:
                            return (self.vol_mol_comp[i] /
                                    self.vol_mol_comp[j]) * \
                                exp(-self.tau[i, j])
                        else:
                            return 1

                    self.Gij_coeff_bubble = \
                        Expression(self._params.component_list,
                                   self._params.component_list,
                                   rule=rule_Gij_coeff_bubble)

                    def rule_A_bubble(self, i):
                        value_1 = log(sum(
                                self.mole_frac_comp[j] *
                                self.Gij_coeff_bubble[j, i]
                                for j in self._params.component_list))
                        return value_1
                    self.A_bubble = Expression(self._params.component_list,
                                               rule=rule_A_bubble)

                    def rule_B_bubble(self, i):
                        value = sum((self.mole_frac_comp[j] *
                                     self.Gij_coeff_bubble[i, j] /
                                     sum(self.mole_frac_comp[k] *
                                         self.Gij_coeff_bubble[k, j]
                                         for k in self._params.component_list))
                                    for j in self._params.component_list)
                        return value
                    self.B_bubble = Expression(self._params.component_list,
                                               rule=rule_B_bubble)

                    def rule_activity_coeff_bubble(self, i):
                        return exp(1 - self.A_bubble[i] - self.B_bubble[i])
                    self.activity_coeff_comp_bubble = \
                        Expression(self._params.component_list,
                                   rule=rule_activity_coeff_bubble)

                    return sum(self.mole_frac_comp[i] *
                               self.activity_coeff_comp_bubble[i] *
                               self._p_sat_bubbleT[i]
                               for i in self._params.component_list) - \
                        self.pressure == 0
            self.eq_temperature_bubble = Constraint(rule=rule_temp_bubble)

        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.temperature_bubble)
            self.del_component(self._p_sat_bubbleT)

    def _temperature_dew(self):

        self.temperature_dew = Var(initialize=298.15,
                                   doc="Dew point temperature (K)")

        def rule_psat_dew(m, j):
            return self._params.pressure_critical[j] * \
                exp((self._params.pressure_sat_coeff[j, "A"] *
                    (1 - self.temperature_dew /
                    self._params.temperature_critical[j]) +
                    self._params.pressure_sat_coeff[j, "B"] *
                    (1 - self.temperature_dew /
                    self._params.temperature_critical[j])**1.5 +
                    self._params.pressure_sat_coeff[j, "C"] *
                    (1 - self.temperature_dew /
                    self._params.temperature_critical[j])**3 +
                    self._params.pressure_sat_coeff[j, "D"] *
                    (1 - self.temperature_dew /
                    self._params.temperature_critical[j])**6) /
                    (1 - (1 - self.temperature_dew /
                          self._params.temperature_critical[j])))

        try:
            # Try to build expression
            self._p_sat_dewT = Expression(self._params.component_list,
                                          rule=rule_psat_dew)

            def rule_temp_dew(self):
                if self.config.parameters.config.activity_coeff_model == \
                        "Ideal":
                    return self.pressure * \
                        sum(self.mole_frac_comp[i] /
                            self._p_sat_dewT[i]
                            for i in self._params.component_list) - 1 == 0
                elif self.config.parameters.config.\
                        activity_coeff_model == "NRTL":
                    # NRTL model variables
                    def rule_Gij_coeff_dew(self, i, j):
                        if i != j:
                            return exp(-self.alpha[i, j] * self.tau[i, j])
                        else:
                            return 1

                    self.Gij_coeff_dew = Expression(
                        self._params.component_list,
                        self._params.component_list,
                        rule=rule_Gij_coeff_dew)

                    def rule_A_dew(self, i):
                        value_1 = sum(self.mole_frac_comp[j] *
                                      self.tau[j, i] * self.Gij_coeff_dew[j, i]
                                      for j in self._params.component_list)
                        value_2 = sum(self.mole_frac_comp[k] *
                                      self.Gij_coeff_dew[k, i]
                                      for k in self._params.component_list)
                        return value_1 / value_2
                    self.A_dew = Expression(self._params.component_list,
                                            rule=rule_A_dew)

                    def rule_B_dew(self, i):
                        value = sum(
                                (self.mole_frac_comp[j] *
                                 self.Gij_coeff_dew[i, j] /
                                 sum(self.mole_frac_comp[k] *
                                     self.Gij_coeff_dew[k, j]
                                     for k in self._params.component_list)) *
                                (self.tau[i, j] - sum(
                                    self.mole_frac_comp[m] *
                                    self.tau[m, j] *
                                    self.Gij_coeff_dew[m, j]
                                    for m in self._params.component_list) /
                                 sum(self.mole_frac_comp[k] *
                                     self.Gij_coeff_dew[k, j]
                                     for k in self._params.component_list))
                                for j in self._params.component_list)
                        return value
                    self.B_dew = Expression(self._params.component_list,
                                            rule=rule_B_dew)

                    def rule_activity_coeff_dew(self, i):
                        return exp(self.A_dew[i] + self.B_dew[i])
                    self.activity_coeff_comp_dew = \
                        Expression(self._params.component_list,
                                   rule=rule_activity_coeff_dew)

                    return sum(self.mole_frac_comp[i] *
                               self.pressure /
                               (self.activity_coeff_comp[i] *
                                self._p_sat_dewT[i])
                               for i in self._params.component_list) - 1 == 0
                else:
                    def rule_Gij_coeff_dew(self, i, j):
                        if i != j:
                            return (self.vol_mol_comp[i] /
                                    self.vol_mol_comp[j]) * \
                                exp(-self.tau[i, j])
                        else:
                            return 1

                    self.Gij_coeff_dew = \
                        Expression(self._params.component_list,
                                   self._params.component_list,
                                   rule=rule_Gij_coeff_dew)

                    def rule_A_dew(self, i):
                        value_1 = log(sum(
                                self.mole_frac_comp[j] *
                                self.Gij_coeff_dew[j, i]
                                for j in self._params.component_list))
                        return value_1
                    self.A_dew = Expression(self._params.component_list,
                                            rule=rule_A_dew)

                    def rule_B_dew(self, i):
                        value = sum((self.mole_frac_comp[j] *
                                     self.Gij_coeff_dew[i, j] /
                                     sum(self.mole_frac_comp[k] *
                                         self.Gij_coeff_dew[k, j]
                                         for k in self._params.component_list))
                                    for j in self._params.component_list)
                        return value
                    self.B_dew = Expression(self._params.component_list,
                                            rule=rule_B_dew)

                    def rule_activity_coeff_dew(self, i):
                        return exp(1 - self.A_dew[i] - self.B_dew[i])
                    self.activity_coeff_comp_dew = \
                        Expression(self._params.component_list,
                                   rule=rule_activity_coeff_dew)

                    return sum(self.mole_frac_comp[i] *
                               self.pressure /
                               (self.activity_coeff_comp[i] *
                                self._p_sat_dewT[i])
                               for i in self._params.component_list) - 1 == 0
            self.eq_temperature_dew = Constraint(rule=rule_temp_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.temperature_dew)
            self.del_component(self._p_sat_dewT)

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

import pyomo.environ as pyo

_all__ = [
    "powerlawp5",
    "powerlaw2",
    "powerlaw3",
    "powerlaw4",
    "avrami2",
    "avrami3",
    "avrami4",
    "avrami5",
    "randomnuc",
    "ptompkins",
    "jander",
    "antijander",
    "valensi",
    "parabolic",
    "gb3d",
    "zlt",
    "grain",
    "massact",
    "massactm",
]

# file containing possible reaction mechanisms
# Due to contraints placed in other subroutines any mechanisms defined here can only take
# process data of an anticipated form [indep variables, conc data]

# Mechanisms defined in ripe have a standard input form due
# To restraints placed on the models by rbfopt in ems
# RIPE standard input form is :
# data = [x,T,flow,vol,t] where x is length ns
# data is not a list of list but a continuous list
# Additional options can be specified for user-specified mechanisms


def powerlawp5(*data):
    pd = data[0]
    return 0.5 * (pd ** (0.5 - (1.0 / 0.5)))


def powerlaw2(*data):
    pd = data[0]
    return 2.0 * (pd ** (2.0 - (1.0 / 2.0)))


def powerlaw3(*data):
    pd = data[0]
    return 3.0 * (pd ** (3.0 - (1.0 / 3.0)))


def powerlaw4(*data):
    pd = data[0]
    return 4.0 * (pd ** (4.0 - (1.0 / 4.0)))


def avrami2(*data):
    pd = data[0]
    return 2.0 * (1.0 - pd) * (-1 * pyo.log(1 - pd)) ** (2.0 - (1.0 / 2.0))


def avrami3(*data):
    pd = data[0]
    return 3.0 * (1.0 - pd) * (-1 * pyo.log(1 - pd)) ** (3.0 - (1.0 / 3.0))


def avrami4(*data):
    pd = data[0]
    return 4.0 * (1.0 - pd) * (-1 * pyo.log(1 - pd)) ** (4.0 - (1.0 / 4.0))


def avrami5(*data):
    pd = data[0]
    return 5.0 * (1.0 - pd) * (-1 * pyo.log(1 - pd)) ** (5.0 - (1.0 / 5.0))


def randomnuc(*data):
    pd = data[0]
    return 1.0 - pd


def ptompkins(*data):
    pd = data[0]
    return pd * (1.0 - pd)


def jander(*data):
    pd = data[0]
    return (
        3.0 * ((1.0 - pd) ** (1.0 / 3.0)) * (1.0 / (1.0 + pd) ** ((-1.0 / 3.0) - 1.0))
    )


def antijander(*data):
    pd = data[0]
    return ((3.0 / 2.0) * (1.0 - pd) ** (2.0 / 3.0)) * (
        1.0 / (1.0 + pd) ** ((-1.0 / 3.0) - 1.0)
    )


def valensi(*data):
    pd = data[0]
    return 1.0 / (-1.0 * pyo.log(1.0 - pd))


def parabolic(*data):
    pd = data[0]
    return 1.0 / (2.0 * pd)


def gb3d(*data):
    pd = data[0]
    return (
        (3.0 / 2.0)
        * (1.0 - pd) ** (4.0 / 3.0)
        * 1.0
        / ((1.0 - pd) ** ((-1.0 / 3.0) - 1.0))
    )


def zlt(*data):
    pd = data[0]
    # zhuralev lesohkin tempelman model
    return (3.0 / 2.0) / ((1.0 - pd) ** ((-1.0 / 3.0) - 1.0))


def grain(*data):
    pd = data[0]
    # grain model
    return (1.0 - pd) ** (2.0 / 3.0)


# Mechanisms can be specified such that they require a mechanism
# EMS requires these models do not take the mechanism as an argument


def massactm(data, def_stoich):
    import numpy as np

    ns = len(def_stoich)
    pwr = []
    x = []
    for i in range(ns):
        if def_stoich[i] < 0:
            pwr.append(abs(def_stoich[i]))
            x.append(data[i])
    return np.product(np.power(x, pwr))


def mechperstoich(mech, stoich):
    import numpy as np

    ns = len(stoich)
    # out_mechs = []

    def massact(*data):
        pwr = []
        x = []
        for i in range(ns):
            if stoich[i] < 0:
                pwr.append(abs(stoich[i]))
                x.append(data[i])
        return np.product(np.power(x, pwr))

    def usr_f(*data):
        #        stoich = stoich
        return mech(data, stoich)

    if mech == "massact":
        f = massact
    else:
        f = usr_f

    return f


# This subroutine analyzes inputs to construct kinetic mechanisms
def getmechs(kwargs):
    import numpy as np

    # Analyze mechanisms in call to ripe
    # determine which stoichiometries to apply each mechanism to
    inkeys = kwargs.keys()
    # Analyze stoichiometries for consistency and relevant information
    # Ideally this input is formatted as a list of list, but can be int, float, or list
    if "stoich" not in inkeys:
        stoich = [[1]]
    else:
        stoich = kwargs["stoich"]
        if isinstance(stoich, type(0)) or isinstance(stoich, type(0.0)):
            stoich = [kwargs["stoich"]]
        elif len(np.shape(stoich)) == 1:
            stoich = [stoich]
        else:
            stoich = kwargs["stoich"]
    dshape = np.shape(stoich)
    if dshape == 1:
        stoich = [stoich]
        dshape = np.shape(stoich)
    nstoich = dshape[0]
    # if len(dshape) > 1:
    #     nspec = dshape[1]
    # else:
    #     nspec = 1
    # Initialize list for later use
    sfixlist = []
    mechlist = []
    # Check for specified mechanisms
    # massaction kinetics internal to ripe are used by default
    # Ideally mechanisms is a list of list of the following format
    # [[[list-of-stoich-indexes],[mechanisms]],...[[next],[sets]]
    if "mech" not in kwargs.keys():
        mechanisms = [[range(nstoich), ["massact"]]]
    else:
        dshape = np.shape(kwargs["mechanisms"])
        if len(dshape) == 1:  # and len(kwargs['mechanisms']) == 1:
            mechanisms = [[range(nstoich), kwargs["mechanisms"]]]
        elif len(dshape) == 0:
            mechanisms = [[range(nstoich), [kwargs["mechanisms"]]]]
        else:
            mechanisms = kwargs["mechanisms"]
    # Determine number of mechanisms specified
    ncons = 0
    for i in range(len(mechanisms)):
        # replace token 'all' with range(stoich)
        # Additionally, either element in mechanisms (may) not be list erroniously
        if not isinstance(mechanisms[i][0], (type([]), range)): 
            mechanisms[i][0] = [mechanisms[i][0]]
        if mechanisms[i][0][0] == "all":
            mechanisms[i][0] = range(nstoich)
        # Ensure the mechanism(s) specified are in a list
        if not isinstance(mechanisms[i][1], type([])):
            mechanisms[i][1] = [mechanisms[i][1]]
        for m in mechanisms[i][1]:
            mechlist.append(m)
            sfixlist.append(mechanisms[i][0])
            ncons += len(mechanisms[i][0])
        if len(mechanisms[i]) < 3:
            mechanisms[i].append(False)
    # nm is the number of mechanisms across stoichiometries
    nm = len(mechlist)
    # fix array is used in the pyomo model to fix binaries for reaction not associated with stoichs
    fixarray = np.zeros([nstoich, nm])
    for j in range(nm):
        for st in sfixlist[j]:
            fixarray[st, j] = 1
    # Now that fixarray is constructed simplify mechanisms
    return stoich, mechanisms, fixarray, ncons, mechlist



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test the 'relations' connecting Resource instances,
which spans the modules idaes.dmf.{dmf, resource, resourcedb}
"""
# stdlib
import logging
import sys

# third-party
import pytest

# local
from idaes.dmf import experiment, resource

# for testing
from .util import tmp_dmf  # need for fixture
from .util import init_logging

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


def test_create_relation_in_resource():
    a = resource.Resource()
    b = resource.Resource()
    resource.create_relation_args(a, "contains", b)
    assert len(a.v["relations"]) == 1
    assert len(b.v["relations"]) == 1
    # bad type
    with pytest.raises(TypeError):
        resource.create_relation("foo", "contains", b)


def test_relation_in_experiment(tmp_dmf):
    e1 = experiment.Experiment(tmp_dmf, name="1")
    a = resource.Resource(value={"name": "foo"})
    e1.add(a)
    assert len(a.v["relations"]) == 1
    assert len(e1.v["relations"]) == 1


def test_relation_with_remove(tmp_dmf):
    e1 = experiment.Experiment(tmp_dmf, name="1")
    n, added = 10, []
    for i in range(n):
        a = resource.Resource({"name": "foo"})
        e1.add(a)
        added.append(a)
    assert len(e1.v["relations"]) == n
    # remove, then update e1
    for a in added:
        tmp_dmf.remove(identifier=a.id)
        e1.update()
        # relation to removed 'a' should be gone
        n -= 1
        assert (len(e1.v["relations"])) == n


def test_find_related(tmp_dmf):
    #
    #  r0
    #   | uses
    #   v
    #   r1
    #   | version
    #   v
    #   r2
    #   /\
    #  /  \ derived
    # v   v
    # r3  r4
    #
    r = [resource.Resource({"name": "r{}".format(i)}) for i in range(5)]
    # r3 <-- derived <-- r2 <-- version <-- r1
    cr = resource.create_relation_args  # shortcut
    cr(r[2], resource.PR_DERIVED, r[3])
    cr(r[1], resource.PR_VERSION, r[2])
    # r4 <-- derived <-- r2
    cr(r[2], resource.PR_DERIVED, r[4])
    # r0 -- Uses --> r1
    cr(r[0], resource.PR_USES, r[1])
    # add to dmf
    for i in range(5):
        tmp_dmf.add(r[i])
    # outgoing from r0 should include 1,2,3,4
    names = []
    for d, rr, m in tmp_dmf.find_related(r[0], meta=["name"]):
        names.append(m["name"])
    names.sort()
    assert names == ["r1", "r2", "r3", "r4"]
    # incoming to r4 should include r0, r1, r2
    names = []
    for d, rr, m in tmp_dmf.find_related(r[4], meta=["name"], outgoing=False):
        names.append(m["name"])
    names.sort()
    assert names == ["r0", "r1", "r2"]


def test_circular(tmp_dmf):
    #
    # r0 -> derived -> r1 -> derived >- r2 -+
    #  ^                                    |
    #  +------------------------------------+
    #     uses
    r = [resource.Resource({"name": "r{}".format(i)}) for i in range(3)]
    resource.create_relation_args(r[0], resource.PR_DERIVED, r[1])
    resource.create_relation_args(r[1], resource.PR_DERIVED, r[2])
    resource.create_relation_args(r[2], resource.PR_USES, r[0])
    for rr in r:
        tmp_dmf.add(rr)
    # outgoing from r0
    names = []
    for d, rr, m in tmp_dmf.find_related(r[0], meta=["name"]):
        names.append(m["name"])
    names.sort()
    assert names == ["r0", "r1", "r2"]
    # incoming to r1
    names = []
    for d, rr, m in tmp_dmf.find_related(r[0], meta=["name"], outgoing=False):
        names.append(m["name"])
    names.sort()
    assert names == ["r0", "r1", "r2"]
    # reducing depth shortens output
    names = []
    for d, rr, m in tmp_dmf.find_related(r[0], meta=["name"], maxdepth=2):
        names.append(m["name"])
    names.sort()
    assert names == ["r1", "r2"]
    names = []
    for d, rr, m in tmp_dmf.find_related(r[0], meta=["name"], maxdepth=1):
        names.append(m["name"])
    names.sort()
    assert names == ["r1"]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test tabular data module.

Note: This is an attempt a ~100% test coverage of a module.
"""
# stdlib
import json
import logging
import os
from io import StringIO
import sys

# third-party
import pytest

# package-local
from idaes.dmf import tabular
from idaes.dmf import util, errors

# for testing
from .util import init_logging

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)

F = tabular.Fields

# class: Column

data_column = {
    F.DATA_UNITS: "ms",
    F.DATA_VALUES: [1.0, 2.0, 3.0],
    F.DATA_ERRORS: [0.0, -1.0, -2.0],
    F.DATA_ERRTYPE: "whatever",
    "extra": "ignored",
}

column_name = "hello"


@pytest.fixture
def col():
    return tabular.Column(column_name, data_column)


def test_column(col):
    assert col.name == column_name


def test_column_data(col):
    d = col.data()
    for f in (F.DATA_UNITS, F.DATA_VALUES, F.DATA_ERRORS, F.DATA_ERRTYPE):
        assert d[f] == data_column[f]


def test_column_len(col):
    assert len(col) == len(data_column[F.DATA_VALUES])


# class: Metadata


ex_metadata = {
    F.DTYPE: "foobar",
    F.AUTH: "Philip Roth",
    F.DATE: "1969-01-12",
    F.TITLE: "Portnoy's Complaint",
    F.INFO: "Random House",
}

ex_csv = "\n".join(
    [
        "# This is some metadata",
        "#",
        'Source,Philip Roth,"Portnoy\'s Complaint",Random House,1969-01-12',
        "Notes,Confessions of a lust-ridden, mother-addicted Jewish bachelor",
        "Datatype,foobar",
        "",
        "# All done",
    ]
)


@pytest.fixture
def metadata():
    return tabular.Metadata(ex_metadata)


@pytest.fixture
def csvfile():
    fake_file = StringIO(ex_csv)
    return fake_file


def test_metadata(metadata):
    assert metadata.author == ex_metadata[F.AUTH]
    assert metadata.date == ex_metadata[F.DATE]


def test_metadata_bad():
    inputs = map(StringIO, ["", "Who cares,Some junk", "source,"])
    for inpfile in inputs:
        with pytest.raises(ValueError):
            print('Bad metadata input: "{}"'.format(inpfile.getvalue()))
            tabular.Metadata.from_csv(inpfile)


def test_metadata_good():
    inpfile = StringIO(ex_csv)
    obj = tabular.Metadata.from_csv(inpfile)
    assert obj.author != ""
    assert obj.title != ""
    assert obj.info != ""
    whole_len = len(obj.source)
    parts_len = map(len, [obj.author, obj.title, obj.info])
    assert whole_len > sum(parts_len)


def test_from_csv(csvfile):
    m = tabular.Metadata.from_csv(csvfile)
    assert m.author == ex_metadata[F.AUTH]
    assert m.date == ex_metadata[F.DATE]


def test_md_properties(metadata):
    assert metadata.author is not None
    assert metadata.date is not None
    assert metadata.title is not None
    assert metadata.info is not None
    # setter and getter
    metadata.datatype = "foo"
    assert metadata.datatype == "foo"


# class: TabularData


ex_data = [
    {
        "name": "Density Data",
        "units": "g/cm^3",
        "values": [1.0053, 1.0188, 1.0023],
        "errors": [0.00005, 0.0001, 0.00005],
        "error_type": "absolute",
    }
]

ex_data2 = [
    {
        "name": "Density Data",
        "units": "g/cm^3",
        "values": [1.0053, 1.0188, 1.0023],
        "errors": [0.00005, 0.0001, 0.00005],
        "error_type": "absolute",
    },
    {
        "name": "Density Data2",
        "units": "g/cm^3",
        "values": [1.0053, 1.0188, 1.0023],
        "errors": [0.00005, 0.0001, 0.00005],
        "error_type": "absolute",
    },
]


ex_data2_badlen = [
    {
        "name": "Density Data",
        "units": "g/cm^3",
        "values": [1.0053, 1.0188, 1.0023],
        "errors": [0.00005, 0.0001],
        "error_type": "absolute",
    },
    {
        "name": "Density Data2",
        "units": "g/cm^3",
        "values": [1.0053, 1.0188],
        "errors": [0.00005, 0.0001, 0.00005],
        "error_type": "absolute",
    },
]

ex_data2_csv = [
    "ID,V1 (g/cm^3),Absolute Error,V2 (m/s),Relative Error",
    "1,1.0053, 0, 1.0035, 0.001",
    "2,1.0188, 0, 1.0037, 0.001",
    "3,1.0023, 0, 1.0039, 0.001",
]


ex_data2_csv_short = ["ID", "1"]


@pytest.fixture
def tabdata():
    obj = tabular.TabularData(ex_data)
    return obj


@pytest.fixture
def tabdata2():
    obj = tabular.TabularData(ex_data2)
    return obj


def test_init(tabdata, tabdata2):
    pass


def test_init_badtype():
    with pytest.raises(TypeError):
        tabular.TabularData("foo")
    with pytest.raises(TypeError):
        tabular.TabularData({"foo": "bar"})


def test_init_emptyvalue():
    with pytest.raises(ValueError):
        tabular.TabularData([])


def test_init_badvalue():
    with pytest.raises(ValueError):
        tabular.TabularData([{"foo": "bar"}])


def test_columns(tabdata):
    c = tabdata.columns
    assert c[0]["units"] == ex_data[0]["units"]


def test_len(tabdata, tabdata2):
    assert len(tabdata) == 3
    assert tabdata.num_rows == len(tabdata)
    assert len(tabdata2) == 3
    assert tabdata2.num_rows == len(tabdata2)


def test_names(tabdata):
    assert tabdata.names()[0] == ex_data[0]["name"]


def test_num_columns(tabdata, tabdata2):
    assert tabdata.num_columns == 1
    assert tabdata2.num_columns == 2


def test_get_column(tabdata):
    name = ex_data[0]["name"]
    assert tabdata.get_column(name).values == ex_data[0]["values"]


def test_get_column_bad(tabdata):
    name = "NOPE!"
    with pytest.raises(KeyError):
        tabdata.get_column(name)


def test_get_column_index(tabdata):
    name = ex_data[0]["name"]
    assert tabdata.get_column_index(name) == 0


def test_get_column_index_err(tabdata):
    name = "-no-such-name-"
    with pytest.raises(KeyError):
        tabdata.get_column_index(name)


def test_num_rows(tabdata):
    assert tabdata.num_rows == 3


def test_as_dict(tabdata):
    assert tabdata.as_list() == ex_data


def test_as_arr(tabdata2):
    values, errors = tabdata2.as_arr()
    assert values[0][0] == ex_data2[0][F.DATA_VALUES][0]
    assert values[0][1] == ex_data2[0][F.DATA_VALUES][1]
    assert values[1][0] == ex_data2[1][F.DATA_VALUES][0]
    assert values[1][1] == ex_data2[1][F.DATA_VALUES][1]


def test_td_badlen():
    with pytest.raises(ValueError):
        tabular.TabularData(ex_data2_badlen)


def test_values_dataframe(tabdata2):
    df = tabdata2.values_dataframe()
    name = tabdata2.names()[0]
    values = ex_data2[0][F.DATA_VALUES]
    for i in range(len(values)):
        assert df[name][i] == values[i]


def test_errors_dataframe(tabdata2):
    df = tabdata2.errors_dataframe()
    name = tabdata2.names()[0]
    values = ex_data2[0][F.DATA_ERRORS]
    for i in range(len(values)):
        assert df[name][i] == values[i]


def test_td_dataframe_nopandas(tabdata):
    tabular.pd = None
    with pytest.raises(ImportError):
        df = tabdata.values_dataframe()


def test_td_from_csv():
    strfile = StringIO("\n".join(ex_data2_csv))
    tabular.TabularData.from_csv(strfile)


def test_td_from_csv_badheader():
    csv = ex_data2_csv[:]
    csv[0] += ",Extra"
    strfile = StringIO("\n".join(csv))
    with pytest.raises(ValueError):
        tabular.TabularData.from_csv(strfile)


def test_td_from_csv_badrow():
    csv = ex_data2_csv[:]
    csv[1] += ",1"
    strfile = StringIO("\n".join(csv))
    with pytest.raises(ValueError):
        tabular.TabularData.from_csv(strfile)


def test_td_csv_short():
    infile = StringIO("\n".join(ex_data2_csv_short))
    with pytest.raises(ValueError):
        tabular.TabularData.from_csv(infile)


# class: Table


@pytest.fixture
def table(tabdata, metadata):
    return tabular.Table(data=tabdata, metadata=metadata)


def test_init_table(table):
    pass


def test_table_init_variations():
    tabular.Table(data=ex_data2, metadata=ex_metadata)
    tabular.Table(data=ex_data2, metadata=ex_metadata)
    tabular.Table(data=ex_data2, metadata=[ex_metadata])


def test_table_init_bad(tabdata, metadata):
    with pytest.raises(TypeError):
        tbl = tabular.Table(data=None, metadata=metadata)
    with pytest.raises(ValueError):
        tbl = tabular.Table(data=[], metadata=metadata)
    with pytest.raises(TypeError):
        tbl = tabular.Table(data=tabdata, metadata="hello")


def test_iter_table(table):
    for key, value in table:
        assert key in ["meta", "data"]
        assert isinstance(value, list)


def test_table_attrs(table):
    assert table.data is not None
    assert table.metadata is not None


def test_dump_table(table):
    fp = StringIO()
    table.dump(fp)
    assert len(fp.getvalue()) > 50


def test_dumps_table(table):
    assert len(table.dumps()) > 50
    assert str(table) == table.dumps()


def test_roundtrip_table(table):
    fp = StringIO()
    table.dump(fp)
    fp.seek(0)
    table2 = tabular.Table.load(fp)
    for m1, m2 in zip(table.metadata, table2.metadata):
        assert m1.as_dict() == m2.as_dict()


def _table_reload(fp):
    """Table reload and check logic."""
    fp.seek(0)
    with pytest.raises(errors.DataFormatError):
        _ = tabular.Table.load(fp)


def test_load_table_missing(tabdata, metadata):
    # ValueError with high-level keys missing
    for delkey in F.DATA, F.META:
        fp = StringIO()
        tbl = tabular.Table(data=tabdata, metadata=metadata)
        tbl_json = tbl.as_dict()
        del tbl_json[delkey]
        json.dump(tbl_json, fp)
        _table_reload(fp)
    # ValueError with metadata keys missing
    for mdelkey in F.DTYPE, F.AUTH, F.INFO:
        fp = StringIO()
        tbl = tabular.Table(data=tabdata, metadata=metadata)
        del tbl.metadata[0]._meta[mdelkey]
        tbl.dump(fp)
        _table_reload(fp)


def test_load_table_nometa(tabdata):
    fp = StringIO()
    tbl = tabular.Table(data=tabdata, metadata=[])
    tbl.dump(fp)
    fp.seek(0)
    tabular.Table.load(fp)


# class: TabularObject

# hmmm.. it's entirely abstract..



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for Python code style.
"""
import logging
import os
from pathlib import Path
import subprocess

_log = logging.getLogger(__name__)


# The most stylish dirs in the project
DIRS = [
    str(p)
    for p in (
        Path("idaes/dmf"),
        # Path("apps/ddm-learning/alamo_python/alamopy"),
        # Path("apps/ddm-learning/ripe_python/ripe"),
    )
]


STYLE_CHECK_CMD = "flake8"


def test_flake8():
    cwd = os.getcwd()
    for d in DIRS:
        path = os.path.join(cwd, d)
        if not os.path.exists(path):
            _log.warning(
                f"Target path '{d}' not found in current dir, '{cwd}'. " "Skipping test"
            )
            continue
        if not os.path.isdir(path):
            _log.warning(
                f"Target path '{d}' in current dir, '{cwd}', is not a directory. "
                "Skipping test"
            )
            continue
        cmd = [STYLE_CHECK_CMD, d]
        _log.info(f"Test code style with command '{' '.join(cmd)}'")
        try:
            proc = subprocess.Popen(cmd)
        except FileNotFoundError:
            _log.warning(
                f"Style checker {STYLE_CHECK_CMD} not found. Skipping style tests"
            )
            break
        proc.wait()
        status = proc.returncode
        assert status == 0, f"Style checker '{STYLE_CHECK_CMD}' had errors for {path}"



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Set the alamo and gams paths here.
"""
import collections as col

debug = {}
data = {}

debug['almloc'] = 'alamo'
debug['gamsloc'] = 'gams'

# Initialize some dictionaries to store data later
# different keys are used to automate the printing of the .alm later
data['opts'] = col.OrderedDict.fromkeys(['ninputs', 'noutputs'])
data['labs'] = dict.fromkeys(
    ['savexlabels', 'savezlabels', 'xlinks', 'zlinks'])
data['lstopts'] = {}
data['set4'] = {}
data['stropts'] = {}

# potential arguments given to doalamo(**kwargs) are listed here
# The data dictionary is used to track arguments that will be passed to alamo
# The different keys 'opts','lstopts','stropts','set4' are used to record
# options of different types (float, list, strings, min/max)
data['pargs'] = {}
data['pargs']['opts'] = list(
    ['modeler', 'solvemip', 'linfcns', 'expfcns', 'logfcns', 'sinfcns',
     'cosfcns', 'sampler', 'maxiter', 'tolmaxerror', 'tolesterror', 'sigma',
     'builder', 'linearerror', 'regularizer', 'deltaterm', 'maxterms',
     'convpen', 'constant', 'cvxbic', 'screener', 'ncvf'])
data['pargs']['lstopts'] = list(
    ['multi2power', 'multi3power', 'ratiopower', 'monomialpower', 'zisint',
     'xmin', 'xmax'])
data['pargs']['stropts'] = list(
    ['simulator', 'almname', 'tracefname', 'almopt', 'gamssolver'])
data['pargs']['set4'] = list(['xmin', 'xmax', 'zmin', 'zmax'])

# The debug dictionary is used to track options utilized by the wrapper
# Some of these, such as file paths, could be changed before installing
debug['pargs'] = list(
    ['savescratch', 'savetrace', 'showalm', 'hardset', 'outkeys',
     'expandoutput', 'cvfun', 'almpath', 'gamspath',
     'hardset', 'simwrap', 'loo', 'lmo', 'mock', 'saveopt', 'savegams', 'savepyfcn'])
debug['savepyfcn'] = True
debug['cvfun'] = False
debug['savescratch'] = False
debug['savetrace'] = False
debug['showalm'] = False
debug['temp'] = 0
# expandoutput includes statistics other than the base model
debug['expandoutput'] = True  # Returns additional information in alamopy call
debug['hardset'] = False  # Force alamo to return a 1-term model
debug['simwrap'] = False  # utilize alamopy.simwrapper 
debug['outkeys'] = False  # return dictionary with output keys

# initialize options and flags not availible to user
debug['validation'] = False
debug['bignum'] = 10**10
debug['traindata'] = True

# assess uncertainty of surrogate model
debug['loo'] = False
debug['lmo'] = -1
debug['mock'] = False
debug['saveopt'] = False  # MENGLE for custom constraints/functions
debug['savegams'] = False

# Initialize trace and .alm names
data['stropts']['tracefname'] = 'trace.trc'
data['stropts']['almname'] = 'temp.alm'
# data['opts']['cvxbic'] = 1
# data['almopts']['solvemip'] = 0
# data['almopts']['sampler'] = 1



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Surrogate modeling helper classes and functions.
This is used to run ALAMO on property data.
"""
# stdlib
import logging
# third-party
import numpy as np
from pandas import DataFrame
# package
from idaes.dmf import resource, propdata
# from idaes.dmf.experiment import Experiment
# alamo
from idaes.dmf import errors
from idaes.surrogate import alamopy

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)

alamo_enabled = alamopy.multos.has_alamo()


class SurrogateModel(object):
    """Run ALAMO to generate surrogate models.

    Automatically track the objects in the DMF.

    Example::

        model = SurrogateModel(dmf, simulator='linsim.py')
        rsrc = dmf.fetch_one(1) # get resource ID 1
        data = rsrc.property_table.data
        model.set_input_data(data, ['temp'], 'density')
        results = model.run()
    """

    PARAM_DATA_KEY = 'parameters'  #: Key in resource 'data' for params

    def __init__(self, experiment, **kwargs):
        """Create surrogate model generator.

        This invokes `alamopy.doalamo()` when the `model` attribute
        is retrieved. Subsequent calls return the model that was generated
        by the first call.

        Args:
            experiment (Experiment): Associated parent resource.
            **kwargs: Keyword arguments passed to doalamo()
        Raises:
            errors.AlamoDisabledError: if alamopy cannot be imported or used.
        """
        if not alamo_enabled:
            raise errors.AlamoDisabledError()
        self._x, self._z, self._xv, self._zv = None, None, None, None
        self._kwargs = kwargs
        self._exp = experiment
        self._rsrc = self._create_resource()
        self._exp.add(self._rsrc)

    def set_input_data(self, data, x_colnames, z_colname):
        """Set input from provided dataframe or property data.

        Args:
            data (PropertyData|pandas.DataFrame): Input data
            x_colnames (List[str]|str): One or more column names for parameters
            z_colname (str): Column for response variable
        Returns:
            None
        Raises:
            KeyError: if columns are not found in data
        """
        df = self._get_df(data)
        self._x = np.array(df[x_colnames])
        self._z = np.array(df[z_colname])
        self._add_data(data, dtype='input', x_colnames=x_colnames,
                       z_colname=z_colname)
        if 'xlabels' not in self._kwargs:
            self._kwargs['xlabels'] = x_colnames

    def set_input_data_np(self, x, z, xlabels=None, zlabel='z'):
        """Set input data from numpy arrays.

        Args:
            x (arr): Numpy array with parameters
            xlabels (List[str]): List of labels for x
            zlabel (str): Label for z
            z (arr): Numpy array with response variables

        Returns:
            None
        """
        self._x, self._z = x, z
        xlabels = self._get_xlabels(x, xlabels)
        self._add_data(self._make_dataframe(x, z, xlabels, zlabel),
                       dtype='input', x_colnames=xlabels, z_colname=zlabel)

    def set_validation_data(self, data, x_colnames, z_colname):
        """Set validation data from provided data.

        Args:
            data (PropertyData|pandas.DataFrame): Input data
            x_colnames (List[str]|str): One or more column names for parameters
            z_colname (str): Column for response variable
        Returns:
            None
        Raises:
            KeyError: if columns are not found in data
        """
        df = self._get_df(data)
        self._xv = np.array(df[x_colnames])
        self._zv = np.array(df[z_colname])
        self._add_data(data, dtype='validation', x_colnames=x_colnames,
                       z_colname=z_colname)

    def set_validation_data_np(self, x, z, xlabels=None, zlabel='z'):
        """Set input data from numpy arrays.

        Args:
            x (arr): Numpy array with parameters
            xlabels (List[str]): List of labels for x
            zlabel (str): Label for z
            z (arr): Numpy array with response variables

        Returns:
            None
        """
        self._xv, self._zv = x, z
        xlabels = self._get_xlabels(x, xlabels)
        df = self._make_dataframe(x, z, xlabels, zlabel)
        self._add_data(df, dtype='validation', x_colnames=xlabels,
                       z_colname=zlabel)

    @staticmethod
    def _get_xlabels(x, xlabels):
        if xlabels is None:
            # generate labels x1, x2, .., xN for each column in x
            xlabels = ['x{:d}'.format(i) for i in range(x.shape[1])]
        return xlabels

    @staticmethod
    def _make_dataframe(x, z, xlabels, zlabel=None):
        """Make a dataframe from some numpy inputs.

        Args:
            x (np.Array): Table of x (variables)
            z (np.Array): Vector of z (responses)
            xlabels (List[str]): Names for each x
            zlabel (str): Name for response

        Returns:
            DataFrame: DataFrame with appropriate header
        """
        df = DataFrame(x, columns=xlabels)
        df[zlabel] = z

        return df

    def run(self, **kwargs):
        """Run ALAMO.

        Args:
            **kwargs: Additional arguments merged with those passed to
                      the class constructor. Any duplicate values will
                      override the earlier ones.

        Returns:
            dict: The dictionary returned from :meth:`alamopy.doalamo`
        """
        if kwargs:
            self._kwargs.update(kwargs)
            # update stored resource
            self._rsrc.data = {self.PARAM_DATA_KEY: self._kwargs}
            self._exp.dmf.update(self._rsrc)
        results = self._run_alamo()
        # TODO: add result object to DMF
        return results

    def _run_alamo(self):
        kwargs = self._kwargs.copy()
        if self._xv is not None and self._zv is not None:
            kwargs['xval'], kwargs['zval'] = self._xv, self._zv
        # run alamo to generate model
        try:
            results = alamopy.doalamo(self._x, self._z, **kwargs)
        except alamopy.AlamoError as err:
            raise errors.AlamoError(str(err))
        # return the generated model
        return results

    @staticmethod
    def _get_df(data):
        """Return input, if already a dataframe, or extract one."""
        if isinstance(data, propdata.PropertyData):
            df = data.values_dataframe()
        elif isinstance(data, dict):
            df = DataFrame(data)
        else:
            df = data  # assume dataframe-like object
        return df

    def _add_data(self, data, dtype='', x_colnames=None, z_colname=None,
                  metadata=None):
        """Add the data as a DMF resource.

        Args:
            data (propdata.PropertyData|DataFrame): Data values
            metadata (propdata.Metadata): If present, metadata
            dtype (str): Either 'input' or 'validation', the type of data
            x_colnames (List[str]): List of x (variable) column names
            z_colname (str): z (response) column name
        """
        if isinstance(data, propdata.PropertyData):
            pdata = data.as_arr()
        else:
            # create property data from dataframe
            assert x_colnames is not None
            assert z_colname is not None
            pdata = []
            for col in x_colnames:
                values = list(data[col])
                pdata.append({"name": "{} x:{}".format(dtype, col),
                              "units": "",
                              "values": values,
                              "errors": [0] * len(values),
                              "error_type": "absolute",
                              "type": "property"})
            values = list(data[z_colname])
            pdata.append({'name': '{} z:{}'.format(dtype, z_colname),
                          'units': '',
                          'values': values,
                          'errors': [0] * len(values),
                          'error_type': 'absolute',
                          'type': 'property'})
        # create resource
        r = resource.Resource(type_=resource.TY_PROPERTY)
        r.data = {'data': pdata,
                  'meta': metadata.as_dict() if metadata else {}}
        r.v['aliases'].append(dtype)
        _log.debug('adding resource dtype={}'.format(dtype))
        # add resource to experiment
        self._exp.add(r)
        # add link between resource and the surrogate-model resource
        self._exp.link(r, resource.PR_USES, self._rsrc)

    def _create_resource(self):
        r = resource.Resource(type_=resource.TY_SURRMOD)
        r.v['desc'] = 'SurrogateModel'
        r.data = {self.PARAM_DATA_KEY: self._kwargs}
        return r



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Property package for the hydrodealkylation of toluene to form benzene
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           exp,
                           Expression,
                           Param,
                           PositiveReals,
                           Set,
                           Var)

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        ReactionParameterBlock,
                        ReactionBlockDataBase,
                        ReactionBlockBase)
from idaes.core.util.misc import add_object_reference

# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("HDAReactionParameterBlock")
class HDAReactionParameterData(ReactionParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """
    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(HDAReactionParameterData, self).build()

        self.reaction_block_class = HDAReactionBlock

        # List of valid phases in property package
        self.phase_list = Set(initialize=['Vap'])

        # Component list - a list of component identifiers
        self.component_list = Set(initialize=['benzene',
                                              'toluene',
                                              'hydrogen',
                                              'methane'])

        # Reaction Index
        self.rate_reaction_idx = Set(initialize=["R1"])

        # Reaction Stoichiometry
        self.rate_reaction_stoichiometry = {("R1", "Vap", "benzene"): 1,
                                            ("R1", "Vap", "toluene"): -1,
                                            ("R1", "Vap", "hydrogen"): -1,
                                            ("R1", "Vap", "methane"): 1,
                                            ("R1", "Liq", "benzene"): 0,
                                            ("R1", "Liq", "toluene"): 0,
                                            ("R1", "Liq", "hydrogen"): 0,
                                            ("R1", "Liq", "methane"): 0}

        # Heat of Reaction
        dh_rxn_dict = {"R1": -1.08e5}
        self.dh_rxn = Param(self.rate_reaction_idx,
                            initialize=dh_rxn_dict,
                            doc="Heat of reaction [J/mol]")

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({
                'k_rxn': {'method': '_rate_constant', 'units': 'm^3/mol.s'},
                'reaction_rate': {'method': "_rxn_rate", 'units': 'mol/m^3.s'}
                })
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})


class ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """
    def initialize(blk, outlvl=0, **kwargs):
        '''
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

                     * 0 = no output (default)
                     * 1 = report after each step

        Returns:
            None
        '''
        if outlvl > 0:
            _log.info('{} Initialization Complete.'.format(blk.name))


@declare_process_block_class("HDAReactionBlock",
                             block_class=ReactionBlock)
class HDAReactionBlockData(ReactionBlockDataBase):
    """
    An example reaction package for saponification of ethyl acetate
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(HDAReactionBlockData, self).build()

        # Heat of reaction - no _ref as this is the actual property
        add_object_reference(
                self,
                "dh_rxn",
                self.config.parameters.dh_rxn)

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
from pyomo.environ import SolverFactory


def test_ipopt_available():
    """
    Tries to set-up the IPOPT and returns exception if not available
    """
    if not SolverFactory('ipopt').available():
        raise Exception(
            "Could not find IPOPT. Users are strongly encouraged to have a "
            "version of IPOPT available, as it is the default solver assumed "
            "by many IDAES examples and tests. See the IDAES install "
            "documentation for instructions on how to get IPOPT.")



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example Peng-Robinson parameter block for the VLE calucations for a
benzene-toluene system.

Unless otherwise noted, parameters are from:
"The Properties of Gases and Liquids, 4th Edition", Reid, Prausnitz and Poling,
McGraw-Hill, 1987
"""

# Chages the divide behavior to not do integer division
from __future__ import division

# Import Pyomo libraries
from pyomo.environ import Reals, Param, NonNegativeReals, Set

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.core.util.misc import extract_data

from idaes.property_models.cubic_eos.cubic_prop_pack import (
        CubicParameterData, CubicEoS)
from idaes.logger import getIdaesLogger


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("BTParameterBlock")
class BTParameterData(CubicParameterData):

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(BTParameterData, self).build()

        self.cubic_type = CubicEoS.PR

        self.component_list = Set(initialize=['benzene', 'toluene'])

        # List of components in each phase (optional)
        self.phase_comp = {"Liq": self.component_list,
                           "Vap": self.component_list}

        # List of phase equilibrium index
        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        self.phase_equilibrium_list = \
            {1: ["benzene", ("Vap", "Liq")],
             2: ["toluene", ("Vap", "Liq")]}

        # Thermodynamic reference state
        self.pressure_ref = Param(mutable=True,
                                  default=101325,
                                  doc='Reference pressure [Pa]')
        self.temperature_ref = Param(mutable=True,
                                     default=298.15,
                                     doc='Reference temperature [K]')

        # Gas Constant
        self.gas_const = Param(within=NonNegativeReals,
                               mutable=False,
                               default=8.314462618,
                               doc='Gas Constant [J/mol.K]')

        # Critical Properties
        pressure_crit_data = {'benzene': 48.9e5,
                              'toluene': 41.0e5}

        self.pressure_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(pressure_crit_data),
            doc='Critical pressure [Pa]')

        temperature_crit_data = {'benzene': 562.2,
                                 'toluene': 591.8}

        self.temperature_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(temperature_crit_data),
            doc='Critical temperature [K]')

        # Pitzer acentricity factor
        omega_data = {'benzene': 0.212,
                      'toluene': 0.263}

        self.omega = Param(
            self.component_list,
            within=Reals,
            mutable=False,
            initialize=extract_data(omega_data),
            doc='Acentricity Factor')

        # Peng-Robinson binary interaction parameters
        kappa_data = {
            ('benzene', 'benzene'): 0.0000, ('benzene', 'toluene'): 0.0000,
            ('toluene', 'benzene'): 0.0000, ('toluene', 'toluene'): 0.0000}

        self.kappa = Param(
            self.component_list,
            self.component_list,
            within=Reals,
            mutable=False,
            initialize=extract_data(kappa_data),
            doc='Peng-Robinson binary interaction parameters')

        # Molecular Weights
        mw_comp_data = {'benzene': 78.1136E-3,
                        'toluene': 92.1405E-3}

        self.mw_comp = Param(self.component_list,
                             mutable=False,
                             initialize=extract_data(mw_comp_data),
                             doc="molecular weight kg/mol")

        # Constants for specific heat capacity, enthalpy and entropy
        cp_ig_data = {('benzene', '1'): -3.392E1,
                      ('benzene', '2'): 4.739E-1,
                      ('benzene', '3'): -3.017E-4,
                      ('benzene', '4'): 7.130E-8,
                      ('toluene', '1'): -2.435E1,
                      ('toluene', '2'): 5.125E-1,
                      ('toluene', '3'): -2.765E-4,
                      ('toluene', '4'): 4.911E-8}

        self.cp_ig = Param(self.component_list,
                           ['1', '2', '3', '4'],
                           mutable=False,
                           initialize=extract_data(cp_ig_data),
                           doc="Parameters to compute cp_comp")

        # Standard heats of formation
        # Source: NIST Webbook, https://webbook.nist.gov
        # Retrieved 25th September 2019
        dh_form_data = {'benzene': 82.9e3,
                        'toluene': 50.1e3}

        self.enth_mol_form_ref = Param(self.component_list,
                                       mutable=False,
                                       initialize=extract_data(dh_form_data),
                                       doc="Standard heats of formation")

        # Standard entropy of formation
        # Source: Engineering Toolbox, https://www.engineeringtoolbox.com
        # Retrieved 25th September, 2019
        ds_form_data = {'benzene': -269,
                        'toluene': -321}

        self.entr_mol_form_ref = Param(self.component_list,
                                       mutable=False,
                                       initialize=extract_data(ds_form_data),
                                       doc="Standard entropy of formation")

        # Antoine coefficients for ideal vapour (units: bar, K)
        # This is needed for initial guesses of bubble and dew points
        antoine_data = {('benzene', '1'): 4.202,
                        ('benzene', '2'): 1322,
                        ('benzene', '3'): -38.56,
                        ('toluene', '1'): 4.216,
                        ('toluene', '2'): 1435,
                        ('toluene', '3'): -43.33}

        self.antoine = Param(self.component_list,
                             ['1', '2', '3'],
                             mutable=False,
                             initialize=extract_data(antoine_data),
                             doc="Antoine Parameters to pressure_sat")



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.dmfbase module
"""
import json
import logging
import os
import sys
import tempfile

# third-party
import pytest

# package
from idaes.dmf import resource
from idaes.dmf import errors
from idaes.dmf.dmfbase import DMFConfig, DMF
from .util import init_logging, tmp_dmf, TempDir

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)

prop_json = [
    {
        "meta": {
            "datatype": "MEA",
            "info": "J. Chem. Eng. Data, 2009, Vol 54, pg. 3096-30100",
            "notes": "r is MEA weight fraction in aqueous soln.",
            "authors": "Amundsen, T.G., Lars, E.O., Eimer, D.A.",
            "title": "Density and Viscosity of Monoethanolamine + .etc.",
            "date": "2009",
        },
        "data": [
            {
                "name": "Viscosity Value",
                "units": "mPa-s",
                "values": [2.6, 6.2],
                "error_type": "absolute",
                "errors": [0.06, 0.004],
                "type": "property",
            },
            {"name": "r", "units": "", "values": [0.2, 1000], "type": "state"},
        ],
    }
]


@pytest.fixture(scope="function")
def tmp_propdata_file(request):
    sdir = getattr(request.module, "scratchdir", "/tmp")
    prop = tmp_propdata()
    tmpf = open(os.path.join(sdir, "resource.json"), "w")
    json.dump(prop, tmpf)
    tmpf.close()
    return tmpf


def tmp_propdata():
    return prop_json[0]


def add_resources(dmf_obj, num=3, **attrs):
    ids = []
    for i in range(num):
        r = resource.Resource(value=attrs, type_="test")
        r.data = {"i": i}
        dmf_obj.add(r)
        ids.append(r.id)
    return ids


# Tests


def test_add_property_data(tmp_dmf, tmp_propdata_file):
    tmpf, prop = tmp_propdata_file, tmp_propdata()
    # Add the resource
    r = resource.Resource(type_="property_data")
    r.set_id()
    r.v["creator"] = {"name": "Dan Gunter"}
    m = prop["meta"]
    work = '{authors}, "{title}". {info}, {date}'.format(**m)
    r.v["sources"].append({"source": work, "date": m["date"]})
    r.data = {"notes": m["notes"]}
    r.v["tags"].append("MEA")
    r.v["datafiles"].append({"path": tmpf.name})
    rid = tmp_dmf.add(r)
    assert rid is not None
    # Retrieve the resource
    r2 = tmp_dmf.fetch_one(rid)
    # Validate the resource
    assert r2.type == "property_data"
    assert "MEA" in r2.v["tags"]
    # Remove the resource
    tmp_dmf.remove(identifier=rid)


def test_property_data_file(tmp_dmf, tmp_propdata_file):
    tmpf, prop = tmp_propdata_file, tmp_propdata()
    # Add the resource
    r = resource.Resource(type_="property_data")
    r.v["datafiles"].append({"path": tmpf.name})
    rid = tmp_dmf.add(r)
    assert rid is not None
    r2 = tmp_dmf.fetch_one(rid)
    path = r2.v["datafiles"][0]["path"]
    f2 = open(path, "r")
    j2 = json.load(f2)
    assert j2 == prop


def test_find_propertydata(tmp_dmf):
    # populate DMF with some property data resources
    pj = prop_json[0]
    n = 10
    for i in range(n):
        pd = resource.Resource(value={"data": pj}, type_=resource.TY_PROPERTY)
        tmp_dmf.add(pd)
    # get them back again
    filter_ = {"type": resource.TY_PROPERTY}
    pdata = list(tmp_dmf.find(filter_dict=filter_))
    assert len(pdata) == n


# def test_dmf_init_minimal():
#    pytest.raises(errors.DMFError, DMF)


def test_dmf_init_strfile():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write("Hello")
        pytest.raises(errors.WorkspaceError, DMF, path=tmpdir)


def test_dmf_init_badfile():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write("Hello: There")
        pytest.raises(errors.WorkspaceError, DMF, path=tmpdir)


def test_dmf_init_logconf():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write(
            """
_id: this-is-a-temporary-config
logging:
    idaes.dmf.dmfbase:
        level: debug
        output: _stderr_
    root:
        output: _stdout_
    dmf:
        output: _stdout_
    .dmf.experiment:
        output: _stdout_
    # equivalent to previous
    idaes.dmf.experiment:
        output: /tmp/experiment.log
    # user
    crazy.little.logger:
        level: error
        output: _stderr_
        """
        )
        d = DMF(path=tmpdir)


def test_dmf_init_logconf_badlevel():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write(
            """
_id: this-is-a-temporary-config
logging:
    root:
        level: debug
    idaes.dmf.util:
        level: "This is not a valid level"
        output: _stderr_
        """
        )
        pytest.raises(errors.DMFError, DMF, path=tmpdir)


def test_dmf_init_logconf_badfile():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write(
            """
_id: this-is-a-temporary-config
logging:
    root:
        level: debug
    idaes.dmf.util:
        output: {}
        """.format(
                os.path.join(os.path.sep, *map(str, range(10)))
            )
        )
        pytest.raises(errors.DMFError, DMF, path=tmpdir)


def test_dmf_init_workspace_name():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write(
            "_id: this-is-a-temporary-config"
        )
        d = DMF(path=tmpdir, name="my workspace", desc="It is a great place to work")


def test_dmf_change_traits():
    with TempDir() as tmpdir:
        open(os.path.join(tmpdir, DMF.WORKSPACE_CONFIG), "w").write(
            "_id: this-is-a-temporary-config"
        )
        d = DMF(path=tmpdir, name="my workspace", desc="It is a great place to work")
        assert d.db_file
        d.db_file = "newdb.json"
        assert d.db_file == "newdb.json"


def test_dmf_add(tmp_dmf):
    r = resource.Resource(value={"desc": "test resource"})
    r.do_copy = True  # copy by default
    # (1) Copy, and don't remove {default behavior}
    tmpf1 = tempfile.NamedTemporaryFile(delete=False)
    tmpf1.close()
    r.v["datafiles"].append({"path": tmpf1.name})
    # (2) Copy, and remove original
    tmpf2 = tempfile.NamedTemporaryFile(delete=False)
    tmpf2.close()
    r.v["datafiles"].append({"path": tmpf2.name, "is_tmp": True})
    # (3) Do not copy (or remove)
    tmpf3 = tempfile.NamedTemporaryFile()
    r.v["datafiles"].append({"path": tmpf3.name, "do_copy": False})

    tmp_dmf.add(r)

    os.unlink(tmpf1.name)
    try:
        os.unlink(tmpf2.name)
        assert False, "Expected error"
    except Exception as err:
        pass

    os.unlink(tmpf3.name)
    # This is ignored. It makes no sense to ask the file
    # to be removed, but not copied (just a file delete?!)
    r = resource.Resource(value={"desc": "test resource"})
    r.v["datafiles"].append({"path": "foo", "do_copy": False, "is_tmp": True})
    tmp_dmf.add(r)


def test_dmf_add_duplicate(tmp_dmf):
    r = resource.Resource(value={"desc": "test resource"})
    tmp_dmf.add(r)
    pytest.raises(errors.DuplicateResourceError, tmp_dmf.add, r)


def test_dmf_add_filesystem_err(tmp_dmf):
    r = resource.Resource(value={"desc": "test resource"})
    # create datafile
    tmpf1 = tempfile.NamedTemporaryFile(delete=False)
    tmpf1.close()
    r.v["datafiles"].append({"path": tmpf1.name})
    # now, to get an error, make the DMF datafile path unwritable
    path = os.path.join(tmp_dmf.root, tmp_dmf.datafile_dir)
    os.chmod(path, 000)
    # then try to add the resource, which includes copying the file into
    # the (now unwritable) directory
    pytest.raises(errors.DMFError, tmp_dmf.add, r)
    # make the directory writable again so we can remove it
    os.chmod(path, 0o777)


def test_dmf_add_tmp_no_copy(tmp_dmf):
    r = resource.Resource(value={"desc": "test resource"})
    # create datafile, with temporary-file flag turned on
    tmpdir = tempfile.mkdtemp()
    tmpfile = os.path.join(tmpdir, "foo")
    open(tmpfile, "w")
    r.v["datafiles"].append({"path": tmpfile, "is_tmp": True, "do_copy": True})
    # we want an error trying to COPY this file; to get this,
    # change the permissions of the directory
    os.chmod(tmpdir, 0o400)
    ok = False
    try:
        tmp_dmf.add(r)
    except errors.DMFError:
        ok = True
    finally:
        # change it back and clean up
        os.chmod(tmpdir, 0o700)
        os.unlink(tmpfile)
        os.rmdir(tmpdir)
    if not ok:
        assert False, "DMFError expected"


def test_dmf_add_tmp_no_unlink(tmp_dmf):
    r = resource.Resource(value={"desc": "test resource"})
    # create datafile, with temporary-file flag turned on
    tmpdir = tempfile.mkdtemp()
    tmpfile = os.path.join(tmpdir, "foo")
    open(tmpfile, "w")
    r.v["datafiles"].append({"path": tmpfile, "is_tmp": True, "do_copy": True})
    # we want an error trying to UNLINK this file; to get this,
    # change the permissions of the dir read-only
    os.chmod(tmpdir, 0o500)
    try:
        tmp_dmf.add(r)
    finally:
        # change it back and clean up
        os.chmod(tmpdir, 0o700)
        os.unlink(tmpfile)
        os.rmdir(tmpdir)


def test_dmf_update(tmp_dmf):
    ids = add_resources(tmp_dmf, 2)
    r1 = tmp_dmf.fetch_one(ids[0])
    r1.v[r1.TYPE_FIELD] = "test"
    r1.v["desc"] = "Updated description"
    tmp_dmf.update(r1)
    r1b = tmp_dmf.fetch_one(ids[0])
    assert r1b.v["desc"] == "Updated description"
    r2 = tmp_dmf.fetch_one(ids[1])
    assert r2.v["desc"] != "Updated description"


def test_dmf_update_newtype(tmp_dmf):
    ids = add_resources(tmp_dmf, 1)
    r1 = tmp_dmf.fetch_one(ids[0])
    r1.v[r1.TYPE_FIELD] = "this type is different"
    try:
        tmp_dmf.update(r1)
    except errors.DMFError:
        pass
    else:
        assert False, "DMFError expected for update() with new type"


def test_dmf_remove(tmp_dmf):
    n = 10
    ids = add_resources(tmp_dmf, num=n)
    assert tmp_dmf.count() == n
    while n > 0:
        n = n - 1
        tmp_dmf.remove(ids[n])
        assert tmp_dmf.count() == n


def test_dmf_remove_filter(tmp_dmf):
    n = 10
    ids = add_resources(tmp_dmf, num=n)
    assert tmp_dmf.count() == n
    # remove half of the added resources
    print("@@ remove half")
    tmp_dmf.remove(filter_dict={"data.i": {"$lt": n / 2}})
    n2 = tmp_dmf.count()
    assert n2 == n / 2
    # try to remove the same group (should do nothing
    print("@@ remove more")
    tmp_dmf.remove(filter_dict={"data.i": {"$lt": n / 2}})
    n2 = tmp_dmf.count()
    assert tmp_dmf.count() == n / 2
    # remove the rest
    print("@@ remove the rest")
    tmp_dmf.remove(filter_dict={"data.i": {"$ge": n / 2}})
    assert tmp_dmf.count() == 0


def test_dmf_find(tmp_dmf):
    # populate with batches of records
    # they all have the tag 'all', each batch has 'batch<N>' as well
    # All resources in a batch are given version 1.0.<N>
    # Individual resources will have data of {i: 0..<batchsz-1>}
    batchsz, numbatches = 10, 9
    all_ids = []
    for i in range(numbatches):
        n = batchsz
        batch = "batch{:d}".format(i + 1)
        version = resource.version_list([1, 0, i + 1])
        ids = add_resources(
            tmp_dmf, num=n, tags=["all", batch], version_info={"version": version}
        )
        all_ids.extend(ids)
    if _log.isEnabledFor(logging.DEBUG):
        r = tmp_dmf.fetch_one(all_ids[0])
        _log.debug("First resource:\n{}".format(r))
    # Find all records, 2 ways
    total_num = batchsz * numbatches
    result = list(tmp_dmf.find())
    assert len(result) == total_num
    result = list(tmp_dmf.find({"tags": ["all"]}))
    assert len(result) == total_num
    # Find with 'all'
    result = list(tmp_dmf.find({"tags!": ["all", "batch1"]}))
    assert len(result) == batchsz


def test_dmf_str(tmp_dmf):
    s = str(tmp_dmf)
    assert len(s) > 0


#########################
# DMFConfig             #
#########################


@pytest.fixture
def dmfconfig_tmp():
    """Default file is in user's home directory.
       We don't want to actually modify this with a test.
       So switch it out and switch it back when the fixture
       is done.
    """
    default_filename = DMFConfig._filename
    tmpfile = tempfile.NamedTemporaryFile()
    DMFConfig._filename = tmpfile.name
    yield tmpfile
    tmpfile.close()
    DMFConfig._filename = default_filename


@pytest.fixture
def dmfconfig_none():
    """Default file is in user's home directory.
    Replace it with a nonexistent file.
    """
    default_filename = DMFConfig._filename
    DMFConfig._filename = os.path.join(os.path.sep, "idaes", *map(str, range(20)))
    yield True
    DMFConfig._filename = default_filename


def test_dmfconfig_init_defaults_nofile(dmfconfig_none):
    config = DMFConfig()
    assert config.c == DMFConfig.DEFAULTS


def test_dmfconfig_init_defaults_emptyfile(dmfconfig_tmp):
    config = DMFConfig()
    assert config.c == DMFConfig.DEFAULTS


def test_dmfconfig_init_defaults2(dmfconfig_tmp):
    config = DMFConfig(defaults={"look": "here"})
    assert config.c["look"] == "here"


def test_dmfconfig_bad_file(dmfconfig_tmp):
    dmfconfig_tmp.write(b"{[\n")
    dmfconfig_tmp.file.flush()
    pytest.raises(ValueError, DMFConfig)


def test_dmfconfig_somefile(dmfconfig_tmp):
    dmfconfig_tmp.write(b"workspace: foobar\n")
    dmfconfig_tmp.file.flush()
    config = DMFConfig()


def test_dmfconfig_save(dmfconfig_tmp):
    config = DMFConfig()
    config.save()


def test_dmfconfig_save_nofile(dmfconfig_none):
    config = DMFConfig()
    pytest.raises(IOError, config.save)


def test_dmfconfig_attrs(dmfconfig_tmp):
    config = DMFConfig()
    assert config.workspace is not None




#!/usr/bin/python
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np
import sys

def sixcamel(*x):
    x1,x2 = x
    t1 = np.multiply(4.0-2.1*np.power(x1,2)+np.divide(np.power(x1,4),3.0),np.power(x1,2))
    t2 = np.multiply(4*np.power(x2,2)-4,np.power(x2,2))
    z = t1+np.multiply(x1,x2) + t2
    return z

def ackley(*x):
    import numpy as np
    x1,x2=x
    a = 20
    b = 0.2
    c = 2 * 3.14159
    z = - a  * np.exp(-b*np.sqrt(0.5*(x1**2 + x2**2)))-np.exp(0.5*(np.cos(c*x1)+np.cos(c*x2)))+a+np.exp(1)
    return z

def branin(*x):
    import numpy as np
    x1,x2 = x
    pi = 3.14159
    z = (x2-(5.1/(4*pi**2))*x1**2+(5/pi)*x1-6)**2 + 10*(1-(1/(8*pi))*np.cos(x1)+10)+np.random.normal(0,0.1)
    return z

if __name__ == '__main__':
    sys.stdout.write(' ALAMOpy example functions ')
    sys.stdout.write(' call functions with : ')
    sys.stdout.write(' examples.<name>')
    sys.stdout.write(' <name> = branin ')
    sys.stdout.write('          sixcamel ')
    sys.stdout.write('          ackley ')
    



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import inspect
import pytest
from pyomo.environ import ConcreteModel, Block, Set
from pyomo.common.config import ConfigBlock, ConfigValue
from idaes.core import (ControlVolumeBlockData, CONFIG_Template,
                        MaterialBalanceType, EnergyBalanceType,
                        MomentumBalanceType, FlowDirection,
                        declare_process_block_class,
                        FlowsheetBlockData, UnitModelBlockData, useDefault,
                        PhysicalParameterBlock, ReactionParameterBlock,
                        MaterialFlowBasis, StateBlock, StateBlockData,
                        ReactionBlockBase, ReactionBlockDataBase)
from idaes.core.util.exceptions import (ConfigurationError, DynamicError,
                                        PropertyPackageError, BurntToast)


# -----------------------------------------------------------------------------
# Test Enumerators for balance type options
def test_material_balance_type():
    assert len(MaterialBalanceType) == 6

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        MaterialBalanceType.foo


def test_energy_balance_type():
    assert len(EnergyBalanceType) == 6

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        EnergyBalanceType.foo


def test_momentum_balance_type():
    assert len(MomentumBalanceType) == 5

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        MomentumBalanceType.foo


def testflow_direction():
    assert len(FlowDirection) == 2

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        FlowDirection.foo


# -----------------------------------------------------------------------------
# Test CONFIG_Template
def test_CONFIG_Template():
    c = CONFIG_Template()

    assert len(c) == 17

    for i in c:
        if i == "dynamic":
            assert c[i] == useDefault
        elif i == "material_balance_type":
            assert c[i] == MaterialBalanceType.componentPhase
        elif i == "energy_balance_type":
            assert c[i] == EnergyBalanceType.enthalpyTotal
        elif i == "momentum_balance_type":
            assert c[i] == MomentumBalanceType.pressureTotal
        elif i == "property_package":
            assert c[i] == useDefault
        elif i == "reaction_package":
            assert c[i] is None
        elif i in ["property_package_args", "reaction_package_args"]:
            assert isinstance(c[i], ConfigBlock)
            assert len(c[i]) == 0
        else:
            assert c[i] is False


def test_CONFIG_Template_validation_general():
    # No config argument takes a string, float/int or list
    c = CONFIG_Template()

    for i in c:
        with pytest.raises(ValueError):
            c[i] = "foo"
        with pytest.raises(ValueError):
            c[i] = 10.0
        with pytest.raises(ValueError):
            c[i] = [1, 2]


def test_CONFIG_Template_true_false():
    # Check arguments that accept True/False as values
    c = CONFIG_Template()

    for i in c:
        if i not in ["material_balance_type", "energy_balance_type",
                     "momentum_balance_type", "property_package",
                     "reaction_package", "property_package_args",
                     "reaction_package_args"]:
            c[i] = True
            c[i] = False


def test_CONFIG_Template_material_balance_type():
    c = CONFIG_Template()

    for i in MaterialBalanceType:
        c["material_balance_type"] = i


def test_CONFIG_Template_energy_balance_type():
    c = CONFIG_Template()

    for i in EnergyBalanceType:
        c["energy_balance_type"] = i


def test_CONFIG_Template_momentum_balance_type():
    c = CONFIG_Template()

    for i in MomentumBalanceType:
        c["momentum_balance_type"] = i


# -----------------------------------------------------------------------------
# Mockup classes for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class _UnitData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare("property_package",
                   ConfigValue(default=None))
    CONFIG.declare("property_package_args",
                   ConfigValue(default={}))

    def build(self):
        super(_UnitData, self).build()


# -----------------------------------------------------------------------------
# Testing ControlVolumeBlockData
@declare_process_block_class("CVFrame")
class CVFrameData(ControlVolumeBlockData):
    def build(self):
        super(ControlVolumeBlockData, self).build()


def test_config_block():
    cv = CVFrame(concrete=True)

    assert len(cv.config) == 7
    assert cv.config.dynamic == useDefault
    assert cv.config.has_holdup is useDefault
    assert cv.config.property_package == useDefault
    assert isinstance(cv.config.property_package_args, ConfigBlock)
    assert len(cv.config.property_package_args) == 0
    assert cv.config.reaction_package is None
    assert isinstance(cv.config.reaction_package_args, ConfigBlock)
    assert len(cv.config.reaction_package_args) == 0
    assert cv.config.auto_construct is False


# -----------------------------------------------------------------------------
# Test _setup_dynamics
def test_setup_dynamics_use_parent_value():
    # Test that dynamic = None works correctly
    m = ConcreteModel()
    m.fs = Flowsheet(default={"dynamic": False})
    m.fs.u = Unit(default={"dynamic": False})
    m.fs.u.cv = CVFrame()

    m.fs.u.cv._setup_dynamics()

    assert m.fs.u.cv.config.dynamic is False
    assert m.fs.u.cv.config.has_holdup is False


def test_setup_dynamics_use_parent_value_fail_no_dynamic():
    # Test that default falls back to flowsheet
    fs = Flowsheet(default={"dynamic": False}, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    fs.b.cv = CVFrame()
    fs.b.cv._setup_dynamics()

    assert fs.b.cv.config.dynamic is False


def test_setup_dynamics_dynamic_in_ss():
    # Test that dynamic = None works correctly
    fs = Flowsheet(default={"dynamic": False}, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    # Add a time attribute to make sure the correct failure triggers
    fs.b.time_ref = Set(initialize=[0])

    fs.b.cv = CVFrame(default={"dynamic": True, "has_holdup": True})

    # _setup_dynamics should return DynamicError
    with pytest.raises(DynamicError):
        fs.b.cv._setup_dynamics()


def test_setup_dynamics_dynamic_holdup_inconsistent():
    # Test that dynamic = None works correctly
    fs = Flowsheet(default={"dynamic": True}, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    # Add a time attribute to make sure the correct failure triggers
    fs.b.time_ref = Set(initialize=[0])

    fs.b.cv = CVFrame(default={"dynamic": True, "has_holdup": False})

    # _setup_dynamics should return ConfigurationError
    with pytest.raises(ConfigurationError):
        fs.b.cv._setup_dynamics()


# -----------------------------------------------------------------------------
# Test _get_property_package
@declare_process_block_class("PropertyParameterBlock")
class _PropertyParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PropertyParameterBlock, self).build()

        frm = inspect.stack()[1]
        self._package_module = inspect.getmodule(frm[0])

        self.phase_list = Set(initialize=["p1", "p2"])
        self.component_list = Set(initialize=["c1", "c2"])


def test_get_property_package_set():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock()
    m.cv = CVFrame(default={"property_package": m.pp})
    m.cv._get_property_package()


def test_get_property_package_default_args():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock(
                default={"default_arguments": {"test": "foo"}})
    m.cv = CVFrame(default={"property_package": m.pp})
    m.cv._get_property_package()

    assert m.cv.config.property_package_args["test"] == "foo"


def test_get_reaction_package_module_combine_args():
    # Test that local and default args combine correctly
    m = ConcreteModel()
    m.pp = PropertyParameterBlock(
            default={"default_arguments": {"test1": "foo",
                                           "test2": "bar"}})
    m.cv = CVFrame(default={"property_package": m.pp,
                            "property_package_args": {"test2": "baz",
                                                      "test3": "bar"}})
    m.cv._get_property_package()

    assert m.cv.config.property_package_args["test1"] == "foo"
    assert m.cv.config.property_package_args["test2"] == "baz"
    assert m.cv.config.property_package_args["test3"] == "bar"


# -----------------------------------------------------------------------------
# Test _get_default_prop_pack
def test_get_default_prop_pack_works():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.pp

    m.fs.cv = CVFrame()
    assert m.fs.cv._get_default_prop_pack() == m.fs.pp


# TODO : should test more failure modes
def test_get_default_prop_pack_no_default():
    m = ConcreteModel()
    m.fs = Flowsheet()

    m.fs.cv = CVFrame()
    with pytest.raises(ConfigurationError):
        m.fs.cv._get_default_prop_pack()


def test_get_property_package_call_to_get_default_prop_pack():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.pp

    m.fs.cv = CVFrame()
    m.fs.cv._get_property_package()
    assert m.fs.cv.config.property_package == m.fs.pp


# -----------------------------------------------------------------------------
# Test _get_indexing_sets
def test_get_indexing_sets_missing_phase_list():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock()
    m.pp.del_component(m.pp.phase_list)
    m.cv = CVFrame(default={"property_package": m.pp})
    m.cv._get_property_package()

    with pytest.raises(PropertyPackageError):
        m.cv._get_indexing_sets()


def test_get_indexing_sets_missing_component_list():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock()
    m.pp.del_component(m.pp.component_list)
    m.cv = CVFrame(default={"property_package": m.pp})
    m.cv._get_property_package()

    with pytest.raises(PropertyPackageError):
        m.cv._get_indexing_sets()


# -----------------------------------------------------------------------------
# Test _get_reaction_package
def test_get_reaction_package_none():
    m = ConcreteModel()
    m.r = CVFrame()

    m.r._get_reaction_package()

    assert hasattr(m.r, "reaction_module") is False


@declare_process_block_class("ReactionParameterTestBlock")
class _ReactionParameterBlock(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

        frm = inspect.stack()[1]
        self._package_module = inspect.getmodule(frm[0])


def test_get_reaction_package_module():
    m = ConcreteModel()
    m.rp = ReactionParameterTestBlock(
                default={"default_arguments": {"test": "foo"}})
    m.cv = CVFrame(default={"reaction_package": m.rp})

    m.cv._get_reaction_package()

    assert m.cv.config.reaction_package == m.rp
    assert m.cv.config.reaction_package_args["test"] == "foo"


def test_get_reaction_package_module_default_args():
    # Test that local and default args combine correctly
    m = ConcreteModel()
    m.rp = ReactionParameterTestBlock(
            default={"default_arguments": {"test1": "foo",
                                           "test2": "bar"}})
    m.cv = CVFrame(default={"reaction_package": m.rp,
                            "reaction_package_args": {"test2": "baz",
                                                      "test3": "bar"}})

    m.cv._get_reaction_package()

    assert m.cv.config.reaction_package_args["test1"] == "foo"
    assert m.cv.config.reaction_package_args["test2"] == "baz"
    assert m.cv.config.reaction_package_args["test3"] == "bar"


# -----------------------------------------------------------------------------
# Test build and auto_construct methods
def test_build():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.cv = CVFrame(default={"property_package": m.fs.pp})

    super(CVFrameData, m.fs.cv).build()


def test_add_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.cv = CVFrame()

    with pytest.raises(NotImplementedError):
        m.fs.cv.add_geometry()


def test_auto_construct():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.cv = CVFrame(default={"property_package": m.fs.pp,
                               "auto_construct": True})

    with pytest.raises(NotImplementedError):
        super(CVFrameData, m.fs.cv).build()


# -----------------------------------------------------------------------------
# Test NotImplementedErrors for all property and balance type methods
def test_add_state_blocks():
    m = ConcreteModel()
    m.cv = CVFrame()

    with pytest.raises(NotImplementedError):
        m.cv.add_state_blocks()


def test_add_reaction_blocks():
    m = ConcreteModel()
    m.cv = CVFrame()

    with pytest.raises(NotImplementedError):
        m.cv.add_reaction_blocks()


def test_add_material_balances():
    m = ConcreteModel()
    m.cv = CVFrame()

    for t in MaterialBalanceType:
        if t == MaterialBalanceType.none:
            assert m.cv.add_material_balances(t) is None
        elif t == MaterialBalanceType.useDefault:
            with pytest.raises(ConfigurationError):
                m.cv.add_material_balances(t)
        else:
            with pytest.raises(NotImplementedError):
                m.cv.add_material_balances(t)


def test_add_energy_balances():
    m = ConcreteModel()
    m.cv = CVFrame()

    for t in EnergyBalanceType:
        if t == EnergyBalanceType.none:
            assert m.cv.add_energy_balances(t) is None
        elif t == EnergyBalanceType.useDefault:
            with pytest.raises(ConfigurationError):
                m.cv.add_energy_balances(t)
        else:
            with pytest.raises(NotImplementedError):
                m.cv.add_energy_balances(t)


def test_add_momentum_balances():
    m = ConcreteModel()
    m.cv = CVFrame()

    for t in MomentumBalanceType:
        if t == MomentumBalanceType.none:
            assert m.cv.add_momentum_balances(t) is None
        else:
            with pytest.raises(NotImplementedError):
                m.cv.add_momentum_balances(t)
